from __future__ import annotations

from dataclasses import dataclass, field
from inspect import Parameter, currentframe, signature
from types import FunctionType
from typing import _GenericAlias  # type: ignore[attr-defined]
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Literal,
    NoReturn,
    Optional,
    TypeVar,
    Union,
    cast,
    get_type_hints,
    overload,
)

from typing_extensions import ParamSpec, get_args, get_origin

from .declarations import *
from .monkeypatch import monkeypatch_forward_ref
from .runtime import *
from .runtime import class_to_ref

monkeypatch_forward_ref()

__all__ = [
    "Registry",
    "BUILTINS",
    "BaseExpr",
    "Unit",
    "rewrite",
    "eq",
    "panic",
    "let",
    "delete",
    "union",
    "set_",
    "rule",
    "var",
    "vars_",
    "Fact",
    "expr_parts",
]

T = TypeVar("T")
P = ParamSpec("P")
TYPE = TypeVar("TYPE", bound=type)
CALLABLE = TypeVar("CALLABLE", bound=Callable)
EXPR = TypeVar("EXPR", bound="BaseExpr")

# Attributes which are sometimes added to classes by the interpreter or the dataclass decorator, or by ipython.
# We ignore these when inspecting the class.

IGNORED_ATTRIBUTES = {
    "__module__",
    "__doc__",
    "__dict__",
    "__weakref__",
    "__orig_bases__",
}


@dataclass
class Registry:
    """
    A registry holds all the declerations of classes and functions as well as the mapping
    to and from egg names to python names.
    """

    _decls: Declarations = field(default_factory=Declarations)

    def _on_register_sort(self, name: str) -> None:
        """
        Called whenever a sort is registered.
        """
        pass

    def _on_register_function(self, ref: CallableRef, decl: FunctionDecl) -> None:
        """
        Called whenever a function is registered.
        """
        pass

    def _on_register_rewrite(self, decl: RewriteDecl) -> None:
        """
        Called whenever a rewrite is registered.
        """
        pass

    def _on_register_rule(self, decl: RuleDecl) -> None:
        """
        Called whenever a rule is registered.
        """
        pass

    def _on_register_action(self, decl: ActionDecl) -> None:
        pass

    @overload
    def class_(self, *, egg_sort: str) -> Callable[[TYPE], TYPE]:
        ...

    @overload
    def class_(self, cls: TYPE, /) -> TYPE:
        ...

    def class_(self, *args, **kwargs) -> Any:
        """
        Registers a class.
        """
        hint_locals = currentframe().f_back.f_locals  # type: ignore

        if kwargs:
            assert set(kwargs.keys()) == {"egg_sort"}
            return lambda cls: self._class(cls, hint_locals, kwargs["egg_sort"])
        assert len(args) == 1
        return self._class(args[0], hint_locals)

    def _class(
        self,
        cls: type[BaseExpr],
        hint_locals: dict[str, Any],
        egg_sort: Optional[str] = None,
    ) -> RuntimeClass:
        """
        Registers a class.
        """
        cls_name = cls.__name__
        # Get all the methods from the class
        cls_dict: dict[str, Any] = {
            k: v for k, v in cls.__dict__.items() if k not in IGNORED_ATTRIBUTES
        }
        parameters: list[TypeVar] = cls_dict.pop("__parameters__", [])

        # Register class first
        if cls_name in self._decls.classes:
            raise ValueError(f"Class {cls_name} already registered")
        n_type_vars = len(parameters)
        cls_decl = ClassDecl(n_type_vars=n_type_vars)
        self._decls.classes[cls_name] = cls_decl
        self._decls.register_sort(JustTypeRef(cls_name), egg_sort)
        self._on_register_sort(egg_sort or cls_name)

        # The type ref of self is paramterized by the type vars
        slf_type_ref = TypeRefWithVars(
            cls_name, tuple(ClassTypeVarRef(i) for i in range(n_type_vars))
        )

        # Then register each of its methods
        for method_name, method in cls_dict.items():
            is_init = method_name == "__init__"
            # Don't register the init methods for literals, since those don't use the type checking mechanisms
            if is_init and cls_name in LIT_CLASS_NAMES:
                continue
            if isinstance(method, _WrappedMethod):
                fn = method.fn
                egg_fn = method.egg_fn
                cost = method.cost
                default = method.default
                merge = method.merge
            else:
                fn = method
                egg_fn, cost, default, merge = None, None, None, None
            if isinstance(fn, classmethod):
                fn = fn.__func__
                is_classmethod = True
            else:
                # We count __init__ as a classmethod since it is called on the class
                is_classmethod = is_init

            fn_decl = self._generate_function_decl(
                fn,
                hint_locals,
                default,
                cost,
                merge,
                "cls" if is_classmethod and not is_init else slf_type_ref,
                parameters,
                is_init,
                # If this is an i64, use the runtime class for the alias so that i64Like is resolved properly
                # Otherwise, this might be a Map in which case pass in the original cls so that we
                # can do Map[T, V] on it, which is not allowed on the runtime class
                (
                    RuntimeClass(self._decls, "i64") if cls_name == "i64" else cls,
                    cls_name,
                ),
            )
            ref: MethodRef | ClassMethodRef
            if is_classmethod:
                cls_decl.class_methods[method_name] = fn_decl
                ref = ClassMethodRef(cls_name, method_name)
            else:
                cls_decl.methods[method_name] = fn_decl
                ref = MethodRef(cls_name, method_name)
            self._register_callable_ref(egg_fn, ref)
            self._on_register_function(ref, fn_decl)

        # Register != as a method so we can print it as a string
        self._register_callable_ref("!=", MethodRef(cls_name, "__ne__"))
        return RuntimeClass(self._decls, cls_name)

    # We seperate the function and method overloads to make it simpler to know if we are modifying a function or method,
    # So that we can add the functions eagerly to the registry and wait on the methods till we process the class.

    # We have to seperate method/function overloads for those that use the T params and those that don't
    # Otherwise, if you say just pass in `cost` then the T param is inferred as `Nothing` and
    # It will break the typing.
    @overload
    def method(  # type: ignore
        self,
        *,
        egg_fn: Optional[str] = None,
        cost: Optional[int] = None,
        merge: Optional[Callable[[Any, Any], Any]] = None,
    ) -> Callable[[CALLABLE], CALLABLE]:
        ...

    @overload
    def method(
        self,
        *,
        egg_fn: Optional[str] = None,
        cost: Optional[int] = None,
        default: Optional[EXPR] = None,
        merge: Optional[Callable[[EXPR, EXPR], EXPR]] = None,
    ) -> Callable[[Callable[P, EXPR]], Callable[P, EXPR]]:
        ...

    def method(
        self,
        *,
        egg_fn: Optional[str] = None,
        cost: Optional[int] = None,
        default: Optional[EXPR] = None,
        merge: Optional[Callable[[EXPR, EXPR], EXPR]] = None,
    ) -> Callable[[Callable[P, EXPR]], Callable[P, EXPR]]:
        return lambda fn: _WrappedMethod(egg_fn, cost, default, merge, fn)

    @overload
    def function(self, fn: CALLABLE, /) -> CALLABLE:
        ...

    @overload
    def function(  # type: ignore
        self,
        *,
        egg_fn: Optional[str] = None,
        cost: Optional[int] = None,
        merge: Optional[Callable[[Any, Any], Any]] = None,
    ) -> Callable[[CALLABLE], CALLABLE]:
        ...

    @overload
    def function(
        self,
        *,
        egg_fn: Optional[str] = None,
        cost: Optional[int] = None,
        default: Optional[EXPR] = None,
        merge: Optional[Callable[[EXPR, EXPR], EXPR]] = None,
    ) -> Callable[[Callable[P, EXPR]], Callable[P, EXPR]]:
        ...

    def function(self, *args, **kwargs) -> Any:
        """
        Registers a function.
        """
        fn_locals = currentframe().f_back.f_locals  # type: ignore

        # If we have any positional args, then we are calling it directly on a function
        if args:
            assert len(args) == 1
            return self._function(args[0], fn_locals)
        # otherwise, we are passing some keyword args, so save those, and then return a partial
        return lambda fn: self._function(fn, fn_locals, **kwargs)

    def _function(
        self,
        fn: Callable[..., RuntimeExpr],
        hint_locals: dict[str, Any],
        egg_fn: Optional[str] = None,
        cost: Optional[int] = None,
        default: Optional[RuntimeExpr] = None,
        merge: Optional[Callable[[RuntimeExpr, RuntimeExpr], RuntimeExpr]] = None,
    ) -> RuntimeFunction:
        """
        Uncurried version of function decorator
        """
        name = fn.__name__
        if name in self._decls.functions:
            raise ValueError(f"Function {name} already registered")

        # Save function decleartion
        fn_decl = self._generate_function_decl(fn, hint_locals, default, cost, merge)
        self._decls.functions[name] = fn_decl
        # Register it with the egg name
        fn_ref = FunctionRef(name)
        self._register_callable_ref(egg_fn, fn_ref)
        self._on_register_function(fn_ref, fn_decl)
        # Return a runtime function whcich will act like the decleration
        return RuntimeFunction(self._decls, name)

    def _generate_function_decl(
        self,
        fn: Any,
        # Pass in the locals, retrieved from the frame when wrapping,
        # so that we support classes and function defined inside of other functions (which won't show up in the globals)
        hint_locals: dict[str, Any],
        default: Optional[RuntimeExpr],
        cost: Optional[int],
        merge: Optional[Callable[[RuntimeExpr, RuntimeExpr], RuntimeExpr]],
        # The first arg is either cls, for a classmethod, a self type, or none for a function
        first_arg: Literal["cls"] | TypeOrVarRef | None = None,
        cls_typevars: list[TypeVar] = [],
        is_init: bool = False,
        cls_type_and_name: Optional[tuple[type | RuntimeClass, str]] = None,
    ) -> FunctionDecl:
        if not isinstance(fn, FunctionType):
            raise NotImplementedError(
                f"Can only generate function decls for functions not {fn}  {type(fn)}"
            )

        hint_globals = fn.__globals__.copy()

        if cls_type_and_name:
            hint_globals[cls_type_and_name[1]] = cls_type_and_name[0]
        hints = get_type_hints(fn, hint_globals, hint_locals)
        # If this is an init fn use the first arg as the return type
        if is_init:
            if not isinstance(first_arg, (ClassTypeVarRef, TypeRefWithVars)):
                raise ValueError("Init function must have a self type")
            return_type = first_arg
        else:
            return_type = self._resolve_type_annotation(
                hints["return"], cls_typevars, cls_type_and_name
            )

        params = list(signature(fn).parameters.values())
        # Remove first arg if this is a classmethod or a method, since it won't have an annotation
        if first_arg is not None:
            first, *params = params
            if first.annotation != Parameter.empty:
                raise ValueError(
                    f"First arg of a method must not have an annotation, not {first.annotation}"
                )

        for param in params:
            if param.kind != Parameter.POSITIONAL_OR_KEYWORD:
                raise ValueError(
                    f"Can only register functions with positional or keyword args, not {param.kind}"
                )

        arg_types = tuple(
            self._resolve_type_annotation(
                hints[t.name], cls_typevars, cls_type_and_name
            )
            for t in params
        )
        # If the first arg is a self, and this not an __init__ fn, add this as a typeref
        if isinstance(first_arg, (ClassTypeVarRef, TypeRefWithVars)) and not is_init:
            arg_types = (first_arg,) + arg_types

        default_decl = None if default is None else default.__egg_expr__
        merge_decl = (
            None
            if merge is None
            else merge(
                RuntimeExpr(self._decls, return_type.to_just(), VarDecl("old")),
                RuntimeExpr(self._decls, return_type.to_just(), VarDecl("new")),
            ).__egg_expr__
        )
        decl = FunctionDecl(
            return_type=return_type,
            arg_types=arg_types,
            cost=cost,
            default=default_decl,
            merge=merge_decl,
        )
        return decl

    def _register_callable_ref(self, egg_fn: Optional[str], ref: CallableRef) -> None:
        egg_fn = egg_fn or ref.generate_egg_name()
        if ref in self._decls.callable_ref_to_egg_fn:
            raise ValueError(f"Callable ref {ref} already registered")
        self._decls.callable_ref_to_egg_fn[ref] = egg_fn
        self._decls.egg_fn_to_callable_refs[egg_fn].add(ref)

    def _resolve_type_annotation(
        self,
        tp: object,
        cls_typevars: list[TypeVar],
        cls_type_and_name: Optional[tuple[type | RuntimeClass, str]],
    ) -> TypeOrVarRef:
        if isinstance(tp, TypeVar):
            return ClassTypeVarRef(cls_typevars.index(tp))
        # If there is a union, it should be of a literal and another type to allow type promotion
        if get_origin(tp) == Union:
            args = get_args(tp)
            if len(args) != 2:
                raise TypeError("Union types are only supported for type promotion")
            fst, snd = args
            if fst in {int, str}:
                return self._resolve_type_annotation(
                    snd, cls_typevars, cls_type_and_name
                )
            if snd in {int, str}:
                return self._resolve_type_annotation(
                    fst, cls_typevars, cls_type_and_name
                )
            raise TypeError("Union types are only supported for type promotion")

        # If this is the type for the class, use the class name
        if cls_type_and_name and tp == cls_type_and_name[0]:
            return TypeRefWithVars(cls_type_and_name[1])

        # If this is the class for this method and we have a paramaterized class, recurse
        if (
            cls_type_and_name
            and isinstance(tp, _GenericAlias)
            and tp.__origin__ == cls_type_and_name[0]  # type: ignore
        ):
            return TypeRefWithVars(
                cls_type_and_name[1],
                tuple(
                    self._resolve_type_annotation(a, cls_typevars, cls_type_and_name)
                    for a in tp.__args__  # type: ignore
                ),
            )

        if isinstance(tp, (RuntimeClass, RuntimeParamaterizedClass)):
            return class_to_ref(tp).to_var()
        raise TypeError(f"Unexpected type annotation {tp}")

    def register(self, *values: Rewrite | Rule | Action) -> None:
        """
        Registers any number of rewrites or rules.
        """
        for value in values:
            self._register_single(value)

    def _register_single(self, value: Rewrite | Rule | Action) -> None:
        if isinstance(value, Rewrite):
            rewrite_decl = value._to_decl()
            self._decls.rewrites.append(rewrite_decl)
            self._on_register_rewrite(rewrite_decl)
        elif isinstance(value, Rule):
            rule_decl = value._to_decl()
            self._decls.rules.append(rule_decl)
            self._on_register_rule(rule_decl)
        else:
            action_decl = _action_to_decl(value)
            self._decls.actions.append(action_decl)
            self._on_register_action(action_decl)


@dataclass(frozen=True)
class _WrappedMethod(Generic[P, EXPR]):
    """
    Used to wrap a method and store some extra options on it before processing it.
    """

    egg_fn: Optional[str]
    cost: Optional[int]
    default: Optional[EXPR]
    merge: Optional[Callable[[EXPR, EXPR], EXPR]]
    fn: Callable[P, EXPR]

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> EXPR:
        raise NotImplementedError(
            "We should never call a wrapped method. Did you forget to wrap the class?"
        )


class BaseExpr:
    """
    Expression base class, which adds suport for != to all expression types.
    """

    def __ne__(self: EXPR, other_expr: EXPR) -> Unit:  # type: ignore[override, empty-body]
        """
        Compare whether to expressions are not equal.

        :param self: The expression to compare.
        :param other_expr: The other expression to compare to, which must be of the same type.
        :meta public:
        """
        ...

    def __eq__(self, other: NoReturn) -> NoReturn:  # type: ignore[override, empty-body]
        """
        Equality is currently not supported. We only add this method so that
        if you try to use it MyPy will warn you.
        """
        ...


BUILTINS = Registry()


@BUILTINS.class_(egg_sort="unit")
class Unit(BaseExpr):
    """
    The unit type. This is also used to reprsent if a value exists, if it is resolved or not.
    """

    def __init__(self) -> None:
        ...


# We use these builders so that when creating these structures we can type check
# if the arguments are the same type of expression


def rewrite(lhs: EXPR) -> _RewriteBuilder[EXPR]:
    """Rewrite the given expression to a new expression."""
    return _RewriteBuilder(lhs=lhs)


def eq(expr: EXPR) -> _EqBuilder[EXPR]:
    """Check if the given expression is equal to the given value."""
    return _EqBuilder(expr)


def panic(message: str) -> Panic:
    """Raise an error with the given message."""
    return Panic(message)


def let(name: str, expr: BaseExpr) -> Let:
    """Create a let binding."""
    return Let(name, expr)


def delete(expr: BaseExpr) -> Delete:
    """Create a delete expression."""
    return Delete(expr)


def union(lhs: EXPR) -> _UnionBuilder[EXPR]:
    """Create a union of the given expression."""
    return _UnionBuilder(lhs=lhs)


def set_(lhs: EXPR) -> _SetBuilder[EXPR]:
    """Create a set of the given expression."""
    return _SetBuilder(lhs=lhs)


def rule(*facts: Fact) -> _RuleBuilder:
    """Create a rule with the given facts."""
    return _RuleBuilder(facts=facts)


def var(name: str, bound: type[EXPR]) -> EXPR:
    """Create a new variable with the given name and type."""
    return cast(EXPR, _var(name, bound))


def _var(name: str, bound: Any) -> RuntimeExpr:
    """Create a new variable with the given name and type."""
    if not isinstance(bound, (RuntimeClass, RuntimeParamaterizedClass)):
        raise TypeError(f"Unexpected type {type(bound)}")
    return RuntimeExpr(bound.__egg_decls__, class_to_ref(bound), VarDecl(name))


def vars_(names: str, bound: type[EXPR]) -> Iterable[EXPR]:
    """Create variables with the given names and type."""
    for name in names.split(" "):
        yield var(name, bound)


@dataclass
class _RewriteBuilder(Generic[EXPR]):
    lhs: EXPR

    def to(self, rhs: EXPR, *conditions: Fact) -> Rewrite:
        return Rewrite(lhs=self.lhs, rhs=rhs, conditions=list(conditions))

    def __str__(self) -> str:
        return f"rewrite({self.lhs})"


@dataclass
class _EqBuilder(Generic[EXPR]):
    expr: BaseExpr

    def to(self, *exprs: EXPR) -> Eq:
        return Eq([self.expr, *exprs])

    def __str__(self) -> str:
        return f"eq({self.expr})"


@dataclass
class _SetBuilder(Generic[EXPR]):
    lhs: BaseExpr

    def to(self, rhs: EXPR) -> Set:
        return Set(lhs=self.lhs, rhs=rhs)

    def __str__(self) -> str:
        return f"set_({self.lhs})"


@dataclass
class _UnionBuilder(Generic[EXPR]):
    lhs: BaseExpr

    def with_(self, rhs: EXPR) -> Union_:
        return Union_(lhs=self.lhs, rhs=rhs)

    def __str__(self) -> str:
        return f"union({self.lhs})"


@dataclass
class _RuleBuilder:
    facts: tuple[Fact, ...]

    def then(self, *actions: Action) -> Rule:
        return Rule(actions, self.facts)


def expr_parts(expr: BaseExpr) -> tuple[JustTypeRef, ExprDecl]:
    """
    Returns the underlying type and decleration of the expression. Useful for testing structural equality or debugging.

    :rtype: tuple[object, object]
    """
    assert isinstance(expr, RuntimeExpr)
    return expr.__egg_parts__


@dataclass
class Rewrite:
    lhs: BaseExpr
    rhs: BaseExpr
    conditions: list[Fact]

    def __str__(self) -> str:
        args_str = ", ".join(map(str, [self.rhs, *self.conditions]))
        return f"rewrite({self.lhs}).to({args_str})"

    def _to_decl(self) -> RewriteDecl:
        return RewriteDecl(
            expr_parts(self.lhs)[1],
            expr_parts(self.rhs)[1],
            tuple(_fact_to_decl(fact) for fact in self.conditions),
        )


@dataclass
class Eq:
    exprs: list[BaseExpr]

    def __str__(self) -> str:
        first, *rest = self.exprs
        args_str = ", ".join(map(str, rest))
        return f"eq({first}).to({args_str})"

    def _to_decl(self) -> EqDecl:
        return EqDecl(tuple(expr_parts(expr)[1] for expr in self.exprs))


Fact = Union[Unit, Eq]


def _fact_to_decl(fact: Fact) -> FactDecl:
    if isinstance(fact, Eq):
        return fact._to_decl()
    return expr_parts(fact)[1]


@dataclass
class Delete:
    expr: BaseExpr

    def __str__(self) -> str:
        return f"delete({self.expr})"

    def _to_decl(self) -> DeleteDecl:
        decl = expr_parts(self.expr)[1]
        if not isinstance(decl, CallDecl):
            raise ValueError(f"Can only delete calls not {decl}")
        return DeleteDecl(decl)


@dataclass
class Panic:
    message: str

    def __str__(self) -> str:
        return f"panic({self.message})"

    def _to_decl(self) -> PanicDecl:
        return PanicDecl(self.message)


@dataclass
class Union_(Generic[EXPR]):
    lhs: BaseExpr
    rhs: BaseExpr

    def __str__(self) -> str:
        return f"union({self.lhs}).with_({self.rhs})"

    def _to_decl(self) -> UnionDecl:
        return UnionDecl(expr_parts(self.lhs)[1], expr_parts(self.rhs)[1])


@dataclass
class Set:
    lhs: BaseExpr
    rhs: BaseExpr

    def __str__(self) -> str:
        return f"set_({self.lhs}).to({self.rhs})"

    def _to_decl(self) -> SetDecl:
        lhs = expr_parts(self.lhs)[1]
        if not isinstance(lhs, CallDecl):
            raise ValueError(
                f"Can only create a call with a call for the lhs, got {lhs}"
            )
        return SetDecl(lhs, expr_parts(self.rhs)[1])


@dataclass
class Let:
    name: str
    value: BaseExpr

    def __str__(self) -> str:
        return f"let({self.name}, {self.value})"

    def _to_decl(self) -> LetDecl:
        return LetDecl(self.name, expr_parts(self.value)[1])


Action = Union[Let, Set, Delete, Union_, Panic, "BaseExpr"]


def _action_to_decl(action: Action) -> ActionDecl:
    if isinstance(action, BaseExpr):
        return expr_parts(action)[1]
    return action._to_decl()


@dataclass
class Rule:
    header: tuple[Action, ...]
    body: tuple[Fact, ...]

    def _to_decl(self) -> RuleDecl:
        return RuleDecl(
            tuple(_action_to_decl(action) for action in self.header),
            tuple(_fact_to_decl(fact) for fact in self.body),
        )



import sys
import typing

__all__ = ["monkeypatch_forward_ref"]


def monkeypatch_forward_ref():
    """
    Monkeypatch to backport https://github.com/python/cpython/pull/21553.
    Removed recursive gaurd for simplicity
    Can be removed once Pytho 3.8 is no longer supported
    """
    if sys.version_info >= (3, 9):
        return
    typing.ForwardRef._evaluate = _evaluate_monkeypatch  # type: ignore


def _evaluate_monkeypatch(self, globalns, localns):
    if not self.__forward_evaluated__ or localns is not globalns:
        if globalns is None and localns is None:
            globalns = localns = {}
        elif globalns is None:
            globalns = localns
        elif localns is None:
            localns = globalns
        type_ = typing._type_check(  # type: ignore
            eval(self.__forward_code__, globalns, localns),
            "Forward references must evaluate to types.",
            is_argument=self.__forward_is_argument__,
        )
        self.__forward_value__ = typing._eval_type(  # type: ignore
            type_, globalns, localns
        )
        self.__forward_evaluated__ = True
    return self.__forward_value__



# outline for a myst_nb project with sphinx
# build with: sphinx-build -nW --keep-going -b html . ./_build/html

# load extensions
extensions = [
    "myst_nb",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx_gallery.gen_gallery",
]


##
# Sphinx Gallery
# https://sphinx-gallery.github.io/stable/configuration.html#build-pattern
##

sphinx_gallery_conf = {
    # Run all scripts
    "filename_pattern": r".*",
    "examples_dirs": "../python/egg_smol/examples",
    "gallery_dirs": "auto_examples",
    "abort_on_example_error": True,
    "run_stale_examples": True,
}

##
# https://github.com/tox-dev/sphinx-autodoc-typehints#options
##
always_document_param_types = True
# typehints_defaults = "braces"
# typehints_use_signature = True
# typehints_use_signature_return = True

# specify project details
master_doc = "index"
project = "egg-smol Python"


# basic build settings
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "**.ipynb_checkpoints",
    "auto_examples/*.ipynb",
    # "auto_examples/*.md5",
]
nitpicky = True

html_theme = "pydata_sphinx_theme"

# https://pydata-sphinx-theme.readthedocs.io/en/latest/user_guide/header-links.html#fontawesome-icons
html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/metadsl/egg-smol-python",
            "icon": "fa-brands fa-square-github",
            "type": "fontawesome",
        },
    ],
}

# myst_nb default settings

# Custom formats for reading notebook; suffix -> reader
# nb_custom_formats = {}

# Notebook level metadata key for config overrides
# nb_metadata_key = 'mystnb'

# Cell level metadata key for config overrides
# nb_cell_metadata_key = 'mystnb'

# Mapping of kernel name regex to replacement kernel name(applied before execution)
# nb_kernel_rgx_aliases = {}

# Execution mode for notebooks
nb_execution_mode = "force"

# Path to folder for caching notebooks (default: <outdir>)
# nb_execution_cache_path = ''

# Exclude (POSIX) glob patterns for notebooks
# nb_execution_excludepatterns = ()

# Execution timeout (seconds)
# nb_execution_timeout = 30

# Use temporary folder for the execution current working directory
# nb_execution_in_temp = False

# Allow errors during execution
# nb_execution_allow_errors = False

# Raise an exception on failed execution, rather than emitting a warning
# nb_execution_raise_on_error = False

# Print traceback to stderr on execution error
# nb_execution_show_tb = False

# Merge stdout/stderr execution output streams
# nb_merge_streams = False

# The entry point for the execution output render class (in group `myst_nb.output_renderer`)
# nb_render_plugin = 'default'

# Remove code cell source
# nb_remove_code_source = False

# Remove code cell outputs
# nb_remove_code_outputs = False

# Prompt to expand hidden code cell {content|source|outputs}
# nb_code_prompt_show = 'Show code cell {type}'

# Prompt to collapse hidden code cell {content|source|outputs}
# nb_code_prompt_hide = 'Hide code cell {type}'

# Number code cell source lines
# nb_number_source_lines = False

# Overrides for the base render priority of mime types: list of (builder name, mime type, priority)
# nb_mime_priority_overrides = ()

# Behaviour for stderr output
# nb_output_stderr = 'show'

# Pygments lexer applied to stdout/stderr and text/plain outputs
# nb_render_text_lexer = 'myst-ansi'

# Pygments lexer applied to error/traceback outputs
# nb_render_error_lexer = 'ipythontb'

# Options for image outputs (class|alt|height|width|scale|align)
# nb_render_image_options = {}

# Options for figure outputs (classes|name|caption|caption_before)
# nb_render_figure_options = {}

# The format to use for text/markdown rendering
# nb_render_markdown_format = 'commonmark'

# Javascript to be loaded on pages containing ipywidgets
# nb_ipywidgets_js = {'https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js': {'integrity': 'sha256-Ae2Vz/4ePdIu6ZyI/5ZGsYnb+m0JlOmKPjt6XZ9JJkA=', 'crossorigin': 'anonymous'}, 'https://unpkg.com/@jupyter-widgets/html-manager@^0.20.0/dist/embed-amd.js': {'data-jupyter-widgets-cdn': 'https://cdn.jsdelivr.net/npm/', 'crossorigin': 'anonymous'}}



from __future__ import annotations

from typing_extensions import assert_never

from . import bindings


def run_command(egraph: bindings.EGraph, command: bindings._Command) -> None:
    """Run a command on an EGraph."""
    if isinstance(command, bindings.Datatype):
        egraph.declare_sort(command.name, None)
        for variant in command.variants:
            egraph.declare_constructor(variant, command.name)
    elif isinstance(command, bindings.Sort):
        egraph.declare_sort(command.name, (command.presort, command.args))
    elif isinstance(command, bindings.Function):
        egraph.declare_function(command.decl)
    elif isinstance(command, bindings.Define):
        egraph.define(command.name, command.expr, command.cost)
    elif isinstance(command, bindings.RuleCommand):
        egraph.add_rule(command.rule)
    elif isinstance(command, bindings.RewriteCommand):
        egraph.add_rewrite(command.rewrite)
    elif isinstance(command, bindings.ActionCommand):
        egraph.eval_actions(command.action)
    elif isinstance(command, bindings.Run):
        egraph.run_rules(command.length)
    elif isinstance(command, bindings.Extract):
        egraph.extract_expr(command.expr, command.variants)
    elif isinstance(command, bindings.Check):
        egraph.check_fact(command.fact)
    elif isinstance(command, bindings.ClearRules):
        egraph.clear_rules()
    elif isinstance(command, bindings.Clear):
        egraph.clear()
    elif isinstance(command, bindings.Print):
        egraph.print_function(command.name, command.length)
    elif isinstance(command, bindings.PrintSize):
        egraph.print_size(command.name)
    elif isinstance(command, bindings.Input):
        raise NotImplementedError("Input command not implemented")
    elif isinstance(command, bindings.Query):
        raise NotImplementedError("Query command not implemented")
    elif isinstance(command, bindings.Push):
        for _ in range(command.length):
            egraph.push()
    elif isinstance(command, bindings.Pop):
        for _ in range(command.length):
            egraph.pop()
    else:
        assert_never(command)


def parse_and_run(input: str) -> None:
    """Parse a string and run the commands."""
    egraph = bindings.EGraph()
    for command in bindings.parse(input):
        run_command(egraph, command)



"""
Data only descriptions of the components of an egraph and the expressions.
"""

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from typing import Iterable, Optional, Union, cast

from typing_extensions import assert_never

from . import bindings

__all__ = [
    "Declarations",
    "JustTypeRef",
    "ClassTypeVarRef",
    "TypeRefWithVars",
    "TypeOrVarRef",
    "FunctionRef",
    "MethodRef",
    "ClassMethodRef",
    "CallableRef",
    "ConstantRef",
    "ConstantDecl",
    "FunctionDecl",
    "VarDecl",
    "LitType",
    "LitDecl",
    "CallDecl",
    "ExprDecl",
    "tp_and_expr_decl_from_egg",
    "ClassDecl",
    "RewriteDecl",
    "EqDecl",
    "FactDecl",
    "fact_decl_to_egg",
    "RuleDecl",
    "LetDecl",
    "SetDecl",
    "DeleteDecl",
    "UnionDecl",
    "PanicDecl",
    "ActionDecl",
    "action_decl_to_egg",
]
# Special methods which we might want to use as functions
# Mapping to the operator they represent for pretty printing them
# https://docs.python.org/3/reference/datamodel.html
BINARY_METHODS = {
    "__lt__": "<",
    "__le__": "<=",
    "__eq__": "==",
    "__ne__": "!=",
    "__gt__": ">",
    "__ge__": ">=",
    # Numeric
    "__add__": "+",
    "__sub__": "-",
    "__mul__": "*",
    "__matmul__": "@",
    "__truediv__": "/",
    "__floordiv__": "//",
    "__mod__": "%",
    "__divmod__": "divmod",
    "__pow__": "**",
    "__lshift__": "<<",
    "__rshift__": ">>",
    "__and__": "&",
    "__xor__": "^",
    "__or__": "|",
}
UNARY_METHODS = {
    "__pos__": "+",
    "__neg__": "-",
    "__invert__": "~",
}


@dataclass
class Declarations:
    functions: dict[str, FunctionDecl] = field(default_factory=dict)
    classes: dict[str, ClassDecl] = field(default_factory=dict)
    constants: dict[str, ConstantDecl] = field(default_factory=dict)

    # Bidirectional mapping between egg function names and python callable references.
    # Note that there are possibly mutliple callable references for a single egg function name, like `+`
    # for both int and rational classes.
    egg_fn_to_callable_refs: defaultdict[str, set[CallableRef]] = field(
        default_factory=lambda: defaultdict(set)
    )
    callable_ref_to_egg_fn: dict[CallableRef, str] = field(default_factory=dict)

    # Bidirectional mapping between egg sort names and python type references.
    egg_sort_to_type_ref: dict[str, JustTypeRef] = field(default_factory=dict)
    type_ref_to_egg_sort: dict[JustTypeRef, str] = field(default_factory=dict)

    rewrites: list[RewriteDecl] = field(default_factory=list)
    rules: list[RuleDecl] = field(default_factory=list)
    actions: list[ActionDecl] = field(default_factory=list)

    def integrity_check(self) -> None:
        """
        Checks that:
            1. None of the functions and classes  have the same names
            2. Each mapping is bidrectional
        """
        raise NotImplementedError()

    def get_function_decl(self, ref: CallableRef) -> FunctionDecl:
        if isinstance(ref, FunctionRef):
            return self.functions[ref.name]
        elif isinstance(ref, MethodRef):
            return self.classes[ref.class_name].methods[ref.method_name]
        elif isinstance(ref, ClassMethodRef):
            return self.classes[ref.class_name].class_methods[ref.method_name]
        elif isinstance(ref, ConstantRef):
            return self.constants[ref.name].to_function_decl()
        assert_never(ref)

    def register_sort(
        self, type_ref: JustTypeRef, egg_name: Optional[str] = None
    ) -> str:
        egg_name = egg_name or type_ref.generate_egg_name()
        if egg_name in self.egg_sort_to_type_ref:
            raise ValueError(f"Sort {egg_name} is already registered.")
        self.egg_sort_to_type_ref[egg_name] = type_ref
        self.type_ref_to_egg_sort[type_ref] = egg_name
        return egg_name


# Have two different types of type refs, one that can include vars recursively and one that cannot.
# We only use the one with vars for classmethods and methods, and the other one for egg references as
# well as runtime values.
@dataclass(frozen=True)
class JustTypeRef:
    name: str
    args: tuple[JustTypeRef, ...] = ()

    def generate_egg_name(self) -> str:
        """
        Generates an egg sort name for this type reference by linearizing the type.
        """
        if not self.args:
            return self.name
        args = "_".join(a.generate_egg_name() for a in self.args)
        return f"{self.name}__{args}"

    def to_egg(self, decls: Declarations, egraph: bindings.EGraph) -> str:
        if self in decls.type_ref_to_egg_sort:
            return decls.type_ref_to_egg_sort[self]
        elif not self.args:
            raise ValueError(f"Type {self.name} is not registered.")
        # If this is a type with arguments and it is not registered, then we need to register it
        egg_name = decls.register_sort(self)
        arg_sorts = [
            cast(bindings._Expr, bindings.Var(a.to_egg(decls, egraph)))
            for a in self.args
        ]
        egraph.declare_sort(egg_name, (self.name, arg_sorts))
        return egg_name

    def to_var(self) -> TypeRefWithVars:
        return TypeRefWithVars(self.name, tuple(a.to_var() for a in self.args))

    def pretty(self) -> str:
        if not self.args:
            return self.name
        args = ", ".join(a.pretty() for a in self.args)
        return f"{self.name}[{args}]"


@dataclass(frozen=True)
class ClassTypeVarRef:
    """
    A class type variable represents one of the types of the class, if it is a generic
    class.
    """

    index: int

    def to_just(self) -> JustTypeRef:
        raise NotImplementedError("egg-smol does not support generic classes yet.")


@dataclass(frozen=True)
class TypeRefWithVars:
    name: str
    args: tuple[TypeOrVarRef, ...] = ()

    def to_just(self) -> JustTypeRef:
        return JustTypeRef(self.name, tuple(a.to_just() for a in self.args))


TypeOrVarRef = Union[ClassTypeVarRef, TypeRefWithVars]


@dataclass(frozen=True)
class FunctionRef:
    name: str

    def to_egg(self, decls: Declarations) -> str:
        return decls.callable_ref_to_egg_fn[self]

    def generate_egg_name(self) -> str:
        return self.name


@dataclass(frozen=True)
class MethodRef:
    class_name: str
    method_name: str

    def to_egg(self, decls: Declarations) -> str:
        return decls.callable_ref_to_egg_fn[self]

    def generate_egg_name(self) -> str:
        return f"{self.class_name}__{self.method_name}"


@dataclass(frozen=True)
class ClassMethodRef:
    class_name: str
    method_name: str

    def to_egg(self, decls: Declarations) -> str:
        return decls.callable_ref_to_egg_fn[self]

    def generate_egg_name(self) -> str:
        return f"{self.class_name}__{self.method_name}"


@dataclass(frozen=True)
class ConstantRef:
    name: str

    def to_egg(self, decls: Declarations) -> str:
        return decls.callable_ref_to_egg_fn[self]

    def generate_egg_name(self) -> str:
        return self.name


CallableRef = Union[FunctionRef, MethodRef, ClassMethodRef, ConstantRef]


@dataclass(frozen=True)
class ConstantDecl:
    tp: JustTypeRef
    expr: ExprDecl
    cost: Optional[int] = None

    # A constant is compiled to a function that takes no arguments and returns the constant value
    def to_function_decl(self) -> FunctionDecl:
        return FunctionDecl((), self.tp.to_var(), self.cost)


@dataclass(frozen=True)
class FunctionDecl:
    # TODO: Add arg name to arg so can call with keyword arg
    arg_types: tuple[TypeOrVarRef, ...]
    return_type: TypeOrVarRef
    cost: Optional[int] = None
    default: Optional[ExprDecl] = None
    merge: Optional[ExprDecl] = None

    def to_egg(
        self, decls: Declarations, egraph: bindings.EGraph, ref: CallableRef
    ) -> bindings.FunctionDecl:
        return bindings.FunctionDecl(
            decls.callable_ref_to_egg_fn[ref],
            # Remove all vars from the type refs, raising an errory if we find one,
            # since we cannot create egg functions with vars
            bindings.Schema(
                [a.to_just().to_egg(decls, egraph) for a in self.arg_types],
                self.return_type.to_just().to_egg(decls, egraph),
            ),
            self.default.to_egg(decls) if self.default else None,
            self.merge.to_egg(decls) if self.merge else None,
            self.cost,
        )


@dataclass(frozen=True)
class VarDecl:
    name: str

    @classmethod
    def from_egg(cls, var: bindings.Var) -> tuple[JustTypeRef, LitDecl]:
        raise NotImplementedError(
            "Cannot turn var into egg type because typing unknown."
        )

    def to_egg(self, _decls: Declarations) -> bindings.Var:
        return bindings.Var(self.name)

    def pretty(self, **kwargs) -> str:
        return self.name


LitType = Union[int, str, None]


@dataclass(frozen=True)
class LitDecl:
    value: LitType

    @classmethod
    def from_egg(cls, lit: bindings.Lit) -> tuple[JustTypeRef, LitDecl]:
        if isinstance(lit.value, bindings.Int):
            return JustTypeRef("i64"), cls(lit.value.value)
        if isinstance(lit.value, bindings.String):
            return JustTypeRef("String"), cls(lit.value.value)
        elif isinstance(lit.value, bindings.Unit):
            return JustTypeRef("unit"), cls(None)
        assert_never(lit.value)

    def to_egg(self, _decls: Declarations) -> bindings.Lit:
        if self.value is None:
            return bindings.Lit(bindings.Unit())
        if isinstance(self.value, int):
            return bindings.Lit(bindings.Int(self.value))
        if isinstance(self.value, str):
            return bindings.Lit(bindings.String(self.value))
        assert_never(self.value)

    def pretty(self, wrap_lit=True, **kwargs) -> str:
        """
        Returns a string representation of the literal.

        :param wrap_lit: If True, wraps the literal in a call to the literal constructor.
        """
        if self.value is None:
            return "unit()"
        if isinstance(self.value, int):
            return f"i64({self.value})" if wrap_lit else str(self.value)
        if isinstance(self.value, str):
            return f"String({repr(self.value)})" if wrap_lit else repr(self.value)
        assert_never(self.value)


@dataclass(frozen=True)
class CallDecl:
    callable: CallableRef
    args: tuple[ExprDecl, ...] = ()
    # type parameters that were bound to the callable, if it is a classmethod
    bound_tp_params: Optional[tuple[JustTypeRef, ...]] = None

    def __post_init__(self):
        if self.bound_tp_params and not isinstance(self.callable, ClassMethodRef):
            raise ValueError(
                "Cannot bind type parameters to a non-class method callable."
            )

    @classmethod
    def from_egg(
        cls, decls: Declarations, call: bindings.Call
    ) -> tuple[JustTypeRef, CallDecl]:
        from .type_constraint_solver import TypeConstraintSolver

        results = [tp_and_expr_decl_from_egg(decls, a) for a in call.args]
        arg_types = tuple(r[0] for r in results)
        arg_decls = tuple(r[1] for r in results)

        # Find the first callable ref that matches the call
        for callable_ref in decls.egg_fn_to_callable_refs[call.name]:
            tcs = TypeConstraintSolver()
            fn_decl = decls.get_function_decl(callable_ref)
            return_tp = tcs.infer_return_type(
                fn_decl.arg_types, fn_decl.return_type, arg_types
            )
            return return_tp, cls(callable_ref, arg_decls)
        raise ValueError(f"Could not find callable ref for call {call}")

    def to_egg(self, decls: Declarations) -> bindings.Call:
        """Convert a Call to an egg Call."""
        egg_fn = decls.callable_ref_to_egg_fn[self.callable]
        return bindings.Call(egg_fn, [a.to_egg(decls) for a in self.args])

    def pretty(self, parens=True, **kwargs) -> str:
        """
        Pretty print the call.

        :param parens: If true, wrap the call in parens if it is a binary or unary method call.
        """
        ref, args = self.callable, list(self.args)
        if isinstance(ref, FunctionRef):
            fn_str = ref.name
        elif isinstance(ref, ClassMethodRef):
            tp_ref = JustTypeRef(ref.class_name, self.bound_tp_params or ())
            if ref.method_name == "__init__":
                fn_str = tp_ref.pretty()
            else:
                fn_str = f"{tp_ref.pretty()}.{ref.method_name}"
        elif isinstance(ref, MethodRef):
            name = ref.method_name
            slf, *args = args
            if name in UNARY_METHODS:
                return f"{UNARY_METHODS[name]}{slf.pretty()}"
            elif name in BINARY_METHODS:
                assert len(args) == 1
                expr = f"{slf.pretty()} {BINARY_METHODS[name]} {args[0].pretty(wrap_lit=False)}"
                return expr if not parens else f"({expr})"
            elif name == "__getitem__":
                assert len(args) == 1
                return f"{slf.pretty()}[{args[0].pretty(wrap_lit=False)}]"
            elif name == "__call__":
                return f"{slf.pretty()}({', '.join(a.pretty(wrap_lit=False) for a in args)})"
            fn_str = f"{slf.pretty()}.{name}"
        elif isinstance(ref, ConstantRef):
            return ref.name
        else:
            assert_never(ref)
        return f"{fn_str}({', '.join(a.pretty(wrap_lit=False) for a in args)})"


def test_expr_pretty():
    assert VarDecl("x").pretty() == "x"
    assert LitDecl(42).pretty() == "i64(42)"
    assert LitDecl("foo").pretty() == 'String("foo")'
    assert LitDecl(None).pretty() == "unit()"
    assert CallDecl(FunctionRef("foo"), (VarDecl("x"),)).pretty() == "foo(x)"
    assert (
        CallDecl(
            FunctionRef("foo"), (VarDecl("x"), VarDecl("y"), VarDecl("z"))
        ).pretty()
        == "foo(x, y, z)"
    )
    assert (
        CallDecl(MethodRef("foo", "__add__"), (VarDecl("x"), VarDecl("y"))).pretty()
        == "x + y"
    )
    assert (
        CallDecl(MethodRef("foo", "__getitem__"), (VarDecl("x"), VarDecl("y"))).pretty()
        == "x[y]"
    )
    assert (
        CallDecl(
            ClassMethodRef("foo", "__init__"), (VarDecl("x"), VarDecl("y"))
        ).pretty()
        == "foo(x, y)"
    )
    assert (
        CallDecl(ClassMethodRef("foo", "bar"), (VarDecl("x"), VarDecl("y"))).pretty()
        == "foo.bar(x, y)"
    )
    assert (
        CallDecl(MethodRef("foo", "__call__"), (VarDecl("x"), VarDecl("y"))).pretty()
        == "x(y)"
    )
    assert (
        CallDecl(
            ClassMethodRef("Map", "__init__"),
            (),
            (JustTypeRef("i64"), JustTypeRef("unit")),
        ).pretty()
        == "Map[i64, unit]()"
    )


ExprDecl = Union[VarDecl, LitDecl, CallDecl]


def tp_and_expr_decl_from_egg(
    decls: Declarations, expr: bindings._Expr
) -> tuple[JustTypeRef, ExprDecl]:
    if isinstance(expr, bindings.Var):
        return VarDecl.from_egg(expr)
    if isinstance(expr, bindings.Lit):
        return LitDecl.from_egg(expr)
    if isinstance(expr, bindings.Call):
        return CallDecl.from_egg(decls, expr)
    assert_never(expr)


@dataclass
class ClassDecl:
    methods: dict[str, FunctionDecl] = field(default_factory=dict)
    class_methods: dict[str, FunctionDecl] = field(default_factory=dict)
    n_type_vars: int = 0

    def all_refs(self, name: str) -> Iterable[MethodRef | ClassMethodRef]:
        for method_name in self.methods:
            yield MethodRef(name, method_name)
        for method_name in self.class_methods:
            yield ClassMethodRef(name, method_name)


@dataclass(frozen=True)
class RewriteDecl:
    lhs: ExprDecl
    rhs: ExprDecl
    conditions: tuple[FactDecl, ...]

    def to_egg(self, decls: Declarations) -> bindings.Rewrite:
        return bindings.Rewrite(
            self.lhs.to_egg(decls),
            self.rhs.to_egg(decls),
            [fact_decl_to_egg(decls, c) for c in self.conditions],
        )


@dataclass(frozen=True)
class EqDecl:
    exprs: tuple[ExprDecl, ...]

    def to_egg(self, decls: Declarations) -> bindings.Eq:
        return bindings.Eq([e.to_egg(decls) for e in self.exprs])


FactDecl = Union[ExprDecl, EqDecl]


def fact_decl_to_egg(decls: Declarations, fact: FactDecl) -> bindings._Fact:
    if isinstance(fact, EqDecl):
        return fact.to_egg(decls)
    return bindings.Fact(fact.to_egg(decls))


@dataclass(frozen=True)
class RuleDecl:
    head: tuple[ActionDecl, ...]
    body: tuple[FactDecl, ...]

    def to_egg(self, decls: Declarations) -> bindings.Rule:
        return bindings.Rule(
            [action_decl_to_egg(decls, a) for a in self.head],
            [fact_decl_to_egg(decls, f) for f in self.body],
        )


@dataclass(frozen=True)
class LetDecl:
    name: str
    value: ExprDecl

    def to_egg(self, decls: Declarations) -> bindings.Let:
        return bindings.Let(self.name, self.value.to_egg(decls))


@dataclass(frozen=True)
class SetDecl:
    call: CallDecl
    rhs: ExprDecl

    def to_egg(self, decls: Declarations) -> bindings.Set:
        return bindings.Set(
            self.call.callable.to_egg(decls),
            [a.to_egg(decls) for a in self.call.args],
            self.rhs.to_egg(decls),
        )


@dataclass(frozen=True)
class DeleteDecl:
    call: CallDecl

    def to_egg(self, decls: Declarations) -> bindings.Delete:
        return bindings.Delete(
            self.call.callable.to_egg(decls), [a.to_egg(decls) for a in self.call.args]
        )


@dataclass(frozen=True)
class UnionDecl:
    lhs: ExprDecl
    rhs: ExprDecl

    def to_egg(self, decls: Declarations) -> bindings.Union:
        return bindings.Union(self.lhs.to_egg(decls), self.rhs.to_egg(decls))


@dataclass(frozen=True)
class PanicDecl:
    message: str

    def to_egg(self, _decls: Declarations) -> bindings.Panic:
        return bindings.Panic(self.message)


ActionDecl = Union[LetDecl, SetDecl, DeleteDecl, UnionDecl, PanicDecl, ExprDecl]


def action_decl_to_egg(decls: Declarations, action: ActionDecl) -> bindings._Action:
    if isinstance(action, (CallDecl, LitDecl, VarDecl)):
        return bindings.Expr_(action.to_egg(decls))
    return action.to_egg(decls)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Collection

from .declarations import *

__all__ = ["TypeConstraintSolver", "TypeConstraintError"]


class TypeConstraintError(RuntimeError):
    pass


@dataclass
class TypeConstraintSolver:
    """
    Given some typevars and types, solves the constraints to resolve the typevars.
    """

    # Mapping of class typevar index to their types
    _cls_typevar_index_to_type: dict[int, JustTypeRef] = field(default_factory=dict)

    @classmethod
    def from_type_parameters(cls, tps: Collection[JustTypeRef]) -> TypeConstraintSolver:
        """
        Create a TypeInference from a parameterized class, given as a bound typevar.

        Used for a situation like Map[int, str].create()
        """
        cs = cls()
        # For each param in the class, use this as the ith bound class typevar
        for i, tp in enumerate(tps):
            cs._cls_typevar_index_to_type[i] = tp
        return cs

    def infer_return_type(
        self,
        fn_args: Collection[TypeOrVarRef],
        fn_return: TypeOrVarRef,
        args: Collection[JustTypeRef],
    ) -> JustTypeRef:
        # Infer the type of each type variable based on the actual types of the arguments
        self._infer_typevars_zip(fn_args, args)
        # Substitute the type variables with their inferred types
        return self._subtitute_typevars(fn_return)

    def _infer_typevars_zip(
        self, fn_args: Collection[TypeOrVarRef], args: Collection[JustTypeRef]
    ) -> None:
        if len(fn_args) != len(args):
            raise TypeConstraintError(f"Expected {len(fn_args)} args, got {len(args)}")
        for fn_arg, arg in zip(fn_args, args):
            self._infer_typevars(fn_arg, arg)

    def _infer_typevars(self, fn_arg: TypeOrVarRef, arg: JustTypeRef) -> None:
        if isinstance(fn_arg, TypeRefWithVars):
            if fn_arg.name != arg.name:
                raise TypeConstraintError(f"Expected {fn_arg.name}, got {arg.name}")
            self._infer_typevars_zip(fn_arg.args, arg.args)
        elif fn_arg.index not in self._cls_typevar_index_to_type:
            self._cls_typevar_index_to_type[fn_arg.index] = arg
        elif self._cls_typevar_index_to_type[fn_arg.index] != arg:
            raise TypeConstraintError(f"Expected {fn_arg}, got {arg}")

    def _subtitute_typevars(self, tp: TypeOrVarRef) -> JustTypeRef:
        if isinstance(tp, ClassTypeVarRef):
            return self._cls_typevar_index_to_type[tp.index]
        elif isinstance(tp, TypeRefWithVars):
            return JustTypeRef(
                tp.name,
                tuple(self._subtitute_typevars(arg) for arg in tp.args),
            )


def test_type_inference() -> None:
    import pytest

    i64 = TypeRefWithVars("i64")
    unit = TypeRefWithVars("Unit")
    K, V = ClassTypeVarRef(0), ClassTypeVarRef(1)
    map = TypeRefWithVars("Map", (K, V))
    map_i64_unit = TypeRefWithVars("Map", (i64, unit))

    cs = TypeConstraintSolver()
    assert cs.infer_return_type([i64], i64, [i64.to_just()]) == i64.to_just()
    with pytest.raises(TypeError):
        cs.infer_return_type([i64], i64, [unit.to_just()])
    with pytest.raises(TypeError):
        cs.infer_return_type([], i64, [unit.to_just()])

    assert (
        cs.infer_return_type([map, K], V, [map_i64_unit.to_just(), i64.to_just()])
        == unit.to_just()
    )

    with pytest.raises(TypeError):
        cs.infer_return_type([map, K], V, [map_i64_unit.to_just(), unit.to_just()])

    bound_cs = TypeConstraintSolver.from_type_parameters(
        [i64.to_just(), unit.to_just()]
    )
    assert bound_cs.infer_return_type([K], V, [i64.to_just()]) == unit.to_just()

    with pytest.raises(TypeError):
        bound_cs.infer_return_type([K], V, [unit.to_just()])



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field
from typing import Optional, TypeVar, cast

from . import bindings
from .builtins import BUILTINS, BaseExpr
from .declarations import *
from .registry import *
from .registry import _fact_to_decl
from .runtime import *

__all__ = ["EGraph"]

EXPR = TypeVar("EXPR", bound=BaseExpr)


@dataclass
class EGraph(Registry):
    _egraph: bindings.EGraph = field(default_factory=bindings.EGraph)

    # The current declarations which have been pushed to the stack
    _decl_stack: list[Declarations] = field(default_factory=list)

    def __post_init__(self) -> None:
        # Copy the builtin declarations, so we can add to it
        self._decls = deepcopy(BUILTINS._decls)

    def run(self, iterations: int) -> None:
        """
        Run the egraph for a given number of iterations.
        """
        self._egraph.run_rules(iterations)

    def check(self, fact: Fact) -> None:
        """
        Check if a fact is true in the egraph.
        """
        fact_decl = _fact_to_decl(fact)
        fact_egg = fact_decl_to_egg(self._decls, fact_decl)
        return self._egraph.check_fact(fact_egg)

    def extract(self, expr: EXPR) -> EXPR:
        """
        Extract the lowest cost expression from the egraph.
        """
        tp, decl = expr_parts(expr)
        egg_expr = decl.to_egg(self._decls)
        _cost, new_egg_expr, _variants = self._egraph.extract_expr(egg_expr)
        new_tp, new_decl = tp_and_expr_decl_from_egg(self._decls, new_egg_expr)
        if new_tp != tp:
            raise RuntimeError(f"Type mismatch: {tp} != {new_tp}")
        return cast(EXPR, RuntimeExpr(self._decls, tp, new_decl))

    def extract_multiple(self, expr: EXPR, n: int) -> list[EXPR]:
        """
        Extract multiple expressions from the egraph.
        """
        tp, decl = expr_parts(expr)
        egg_expr = decl.to_egg(self._decls)
        _cost, new_egg_expr, variants = self._egraph.extract_expr(
            egg_expr, variants=n + 1
        )
        new_decls = [
            tp_and_expr_decl_from_egg(self._decls, egg_expr)[1]
            for egg_expr in variants[::-1]
        ]
        return [
            cast(EXPR, RuntimeExpr(self._decls, tp, new_decl)) for new_decl in new_decls
        ]

    def define(self, name: str, expr: EXPR, cost: Optional[int] = None) -> EXPR:
        """
        Define a new expression in the egraph and return a reference to it.
        """
        tp, decl = expr_parts(expr)
        self._egraph.define(name, decl.to_egg(self._decls), cost)
        self._decls.constants[name] = ConstantDecl(tp, decl, cost)
        self._register_callable_ref(name, ConstantRef(name))
        return cast(EXPR, RuntimeExpr(self._decls, tp, VarDecl(name)))

    def push(self) -> None:
        """
        Push the current state of the egraph, so that it can be popped later and reverted back.
        """
        self._egraph.push()
        self._decl_stack.append(self._decls)
        self._decls = deepcopy(self._decls)

    def pop(self) -> None:
        """
        Pop the current state of the egraph, reverting back to the previous state.
        """
        self._egraph.pop()
        self._decls = self._decl_stack.pop()

    def __enter__(self):
        """
        Copy the egraph state, so that it can be reverted back to the original state at the end.
        """
        self.push()

    def __exit__(self, exc_type, exc, exc_tb):
        self.pop()

    def _on_register_function(self, ref: CallableRef, decl: FunctionDecl) -> None:
        # Don't need to registry constants, since they are already registered
        if isinstance(ref, ConstantRef):
            raise RuntimeError("Constants should not be registered as functions")
        egg_decl = decl.to_egg(self._decls, self._egraph, ref)
        self._egraph.declare_function(egg_decl)

    def _on_register_sort(self, name: str) -> None:
        self._egraph.declare_sort(name, None)

    def _on_register_rewrite(self, rewrite: RewriteDecl) -> None:
        self._egraph.add_rewrite(rewrite.to_egg(self._decls))

    def _on_register_rule(self, rule: RuleDecl) -> None:
        self._egraph.add_rule(rule.to_egg(self._decls))

    def _on_register_action(self, decl: ActionDecl) -> None:
        self._egraph.eval_actions(action_decl_to_egg(self._decls, decl))



import os

ROOT_DIR = os.path.dirname(__file__)
# So that it finds the local typings
os.environ["MYPYPATH"] = os.path.join(ROOT_DIR, "python")


# Add mypy's test searcher so we can write test files to verify type checking
pytest_plugins = ["mypy.test.data"]
# Set this to the root directory so it finds the `test-data` directory
os.environ["MYPY_TEST_PREFIX"] = ROOT_DIR



"""
Matrix multiplication and Kronecker product.
============================================
"""
from __future__ import annotations

from egg_smol import *

egraph = EGraph()


@egraph.class_
class Dim(BaseExpr):
    """
    A dimension of a matix.

    >>> Dim(3) * Dim.named("n")
    Dim(3) * Dim.named("n")
    """

    @egraph.method(egg_fn="Lit")
    def __init__(self, value: i64Like) -> None:
        ...

    @egraph.method(egg_fn="NamedDim")
    @classmethod
    def named(cls, name: StringLike) -> Dim:  # type: ignore[empty-body]
        ...

    @egraph.method(egg_fn="Times")
    def __mul__(self, other: Dim) -> Dim:  # type: ignore[empty-body]
        ...


a, b, c, n = vars_("a b c n", Dim)
i, j = vars_("i j", i64)
egraph.register(
    rewrite(a * (b * c)).to((a * b) * c),
    rewrite((a * b) * c).to(a * (b * c)),
    rewrite(Dim(i) * Dim(j)).to(Dim(i * j)),
    rewrite(a * b).to(b * a),
)


@egraph.class_(egg_sort="MExpr")
class Matrix(BaseExpr):
    @egraph.method(egg_fn="Id")
    @classmethod
    def identity(cls, dim: Dim) -> Matrix:  # type: ignore[empty-body]
        """
        Create an identity matrix of the given dimension.
        """
        ...

    @egraph.method(egg_fn="NamedMat")
    @classmethod
    def named(cls, name: StringLike) -> Matrix:  # type: ignore[empty-body]
        """
        Create a named matrix.
        """
        ...

    @egraph.method(egg_fn="MMul")
    def __matmul__(self, other: Matrix) -> Matrix:  # type: ignore[empty-body]
        """
        Matrix multiplication.
        """
        ...

    @egraph.method(egg_fn="nrows")
    def nrows(self) -> Dim:  # type: ignore[empty-body]
        """
        Number of rows in the matrix.
        """
        ...

    @egraph.method(egg_fn="ncols")
    def ncols(self) -> Dim:  # type: ignore[empty-body]
        """
        Number of columns in the matrix.
        """
        ...


@egraph.function(egg_fn="Kron")
def kron(a: Matrix, b: Matrix) -> Matrix:  # type: ignore[empty-body]
    """
    Kronecker product of two matrices.

    https://en.wikipedia.org/wiki/Kronecker_product#Definition
    """
    ...


A, B, C, D = vars_("A B C D", Matrix)
egraph.register(
    # The dimensions of a kronecker product are the product of the dimensions
    rewrite(kron(A, B).nrows()).to(A.nrows() * B.nrows()),
    rewrite(kron(A, B).ncols()).to(A.ncols() * B.ncols()),
    # The dimensions of a matrix multiplication are the number of rows of the first
    # matrix and the number of columns of the second matrix.
    rewrite((A @ B).nrows()).to(A.nrows()),
    rewrite((A @ B).ncols()).to(B.ncols()),
    # The dimensions of an identity matrix are the input dimension
    rewrite(Matrix.identity(n).nrows()).to(n),
    rewrite(Matrix.identity(n).ncols()).to(n),
)
egraph.register(
    # Multiplication by an identity matrix is the same as the other matrix
    rewrite(Matrix.identity(n) @ A).to(A),
    rewrite(A @ Matrix.identity(n)).to(A),
    # Matrix multiplication is associative
    rewrite(A @ (B @ C)).to((A @ B) @ C),
    rewrite((A @ B) @ C).to(A @ (B @ C)),
    # Kronecker product is associative
    rewrite(kron(A, kron(B, C))).to(kron(kron(A, B), C)),
    rewrite(kron(kron(A, B), C)).to(kron(A, kron(B, C))),
    # Kronecker product distributes over matrix multiplication
    rewrite(kron(A @ C, B @ D)).to(kron(A, B) @ kron(C, D)),
    rewrite(kron(A, B) @ kron(C, D)).to(
        kron(A @ C, B @ D),
        # Only when the dimensions match
        eq(A.ncols()).to(C.nrows()),
        eq(B.ncols()).to(D.nrows()),
    ),
)
egraph.register(
    # demand rows and columns when we multiply matrices
    rule(eq(C).to(A @ B)).then(
        let("demand1", A.ncols()),
        let("demand2", A.nrows()),
        let("demand3", B.ncols()),
        let("demand4", B.nrows()),
    ),
    # demand rows and columns when we take the kronecker product
    rule(eq(C).to(kron(A, B))).then(
        let("demand1", A.ncols()),
        let("demand2", A.nrows()),
        let("demand3", B.ncols()),
        let("demand4", B.nrows()),
    ),
)


# Define a number of dimensions
n = egraph.define("n", Dim.named("n"))
m = egraph.define("m", Dim.named("m"))
p = egraph.define("p", Dim.named("p"))

# Define a number of matrices
A = egraph.define("A", Matrix.named("A"))
B = egraph.define("B", Matrix.named("B"))
C = egraph.define("C", Matrix.named("C"))

# Set each to be a square matrix of the given dimension
egraph.register(
    set_(A.nrows()).to(n),
    set_(A.ncols()).to(n),
    set_(B.nrows()).to(m),
    set_(B.ncols()).to(m),
    set_(C.nrows()).to(p),
    set_(C.ncols()).to(p),
)
# Create an example which should equal the kronecker product of A and B
ex1 = egraph.define("ex1", kron(Matrix.identity(n), B) @ kron(A, Matrix.identity(m)))
rows = egraph.define("rows", ex1.nrows())
cols = egraph.define("cols", ex1.ncols())

egraph.run(20)

egraph.check(eq(B.nrows()).to(m))
egraph.check(eq(kron(Matrix.identity(n), B).nrows()).to(n * m))

# Verify it matches the expected result
# TODO
simple_ex1 = egraph.define("simple_ex1", kron(A, B))
egraph.check(eq(ex1).to(simple_ex1))

ex2 = egraph.define("ex2", kron(Matrix.identity(p), C) @ kron(A, Matrix.identity(m)))

egraph.run(10)
# Verify it is not simplified
egraph.check(ex2 != kron(A, C))



from __future__ import annotations

import sys

from mypy.test import helpers

# Monkeypatch testfile_pyversion so that we can test against the current python version, instead of the default (3.7)


def monkeypatched_testfile_pyversion(path: str) -> tuple[int, int]:
    return sys.version_info[:2]


helpers.testfile_pyversion = monkeypatched_testfile_pyversion


# This is required using the `TypeCheckSuite`, because it calls remove on a number of builtin files.
# we can either make empty versions of them or just monkypatch the function so that it works.

# Monkeypatch find_test_files so that we can "remove" any files that don't really exist.

original_find_test_files = helpers.find_test_files


class ListRemoveAnything(list):
    def remove(self, item):
        pass


def new_find_test_files(*args, **kwargs):
    original_output = original_find_test_files(*args, **kwargs)
    return ListRemoveAnything(original_output)


helpers.find_test_files = new_find_test_files

# Import TypeCheckSuite so it is picked up by pytest.
from mypy.test.testcheck import TypeCheckSuite  # noqa: E402

__all__ = ["TypeCheckSuite"]



from __future__ import annotations

import importlib
import pathlib

import pytest
from egg_smol import *

EXAMPLE_FILES = list(
    (pathlib.Path(__file__).parent / "../egg_smol/examples").glob("*.py")
)


# Test all files in the `examples` directory by importing them in this parametrized test
@pytest.mark.parametrize(
    "name", [f.stem for f in EXAMPLE_FILES if f.stem != "__init__"]
)
def test_example(name):
    importlib.import_module(f"egg_smol.examples.{name}")


class TestExprStr:
    def test_unwrap_lit(self):
        assert str(i64(1) + 1) == "i64(1) + 1"
        assert str(i64(1).max(2)) == "i64(1).max(2)"


def test_eqsat_basic():
    egraph = EGraph()

    @egraph.class_
    class Math(BaseExpr):
        def __init__(self, value: i64Like) -> None:
            ...

        @classmethod
        def var(cls, v: StringLike) -> Math:  # type: ignore[empty-body]
            ...

        def __add__(self, other: Math) -> Math:  # type: ignore[empty-body]
            ...

        def __mul__(self, other: Math) -> Math:  # type: ignore[empty-body]
            ...

    # expr1 = 2 * (x + 3)
    expr1 = egraph.define("expr1", Math(2) * (Math.var("x") + Math(3)))

    # expr2 = 6 + 2 * x
    expr2 = egraph.define("expr2", Math(6) + Math(2) * Math.var("x"))

    a, b, c = vars_("a b c", Math)
    x, y = vars_("x y", i64)

    egraph.register(
        rewrite(a + b).to(b + a),
        rewrite(a * (b + c)).to((a * b) + (a * c)),
        rewrite(Math(x) + Math(y)).to(Math(x + y)),
        rewrite(Math(x) * Math(y)).to(Math(x * y)),
    )

    egraph.run(10)

    egraph.check(eq(expr1).to(expr2))


def test_fib():
    egraph = EGraph()

    @egraph.function
    def fib(x: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    f0, f1, x = vars_("f0 f1 x", i64)
    egraph.register(
        set_(fib(0)).to(i64(1)),
        set_(fib(1)).to(i64(1)),
        rule(
            eq(f0).to(fib(x)),
            eq(f1).to(fib(x + 1)),
        ).then(set_(fib(x + 2)).to(f0 + f1)),
    )
    egraph.run(7)
    egraph.check(eq(fib(i64(7))).to(i64(21)))


def test_fib_demand():
    egraph = EGraph()

    @egraph.class_
    class Num(BaseExpr):
        def __init__(self, i: i64Like) -> None:
            ...

        def __add__(self, other: Num) -> Num:  # type: ignore[empty-body]
            ...

    @egraph.function
    def fib(x: i64Like) -> Num:  # type: ignore[empty-body]
        ...

    a, b, x = vars_("a b x", i64)
    f = var("f", Num)
    egraph.register(
        rewrite(Num(a) + Num(b)).to(Num(a + b)),
        rule(eq(f).to(fib(x)), x > 1).then(set_(fib(x)).to(fib(x - 1) + fib(x - 2))),
        set_(fib(0)).to(Num(0)),
        set_(fib(1)).to(Num(1)),
    )
    f7 = egraph.define("f7", fib(7), cost=2)
    egraph.run(14)
    egraph.check(eq(f7).to(Num(13)))
    res = egraph.extract(f7)
    assert expr_parts(res) == expr_parts(Num(13))


def test_resolution():
    egraph = EGraph()

    @egraph.class_
    class Bool(BaseExpr):
        @egraph.method(egg_fn="True")
        @classmethod
        def true(cls) -> Bool:  # type: ignore[empty-body]
            ...

        @egraph.method(egg_fn="False")
        @classmethod
        def false(cls) -> Bool:  # type: ignore[empty-body]
            ...

        @egraph.method(egg_fn="or")
        def __or__(self, other: Bool) -> Bool:  # type: ignore[empty-body]
            ...

        @egraph.method(egg_fn="neg")
        def __invert__(self) -> Bool:  # type: ignore[empty-body]
            ...

    T = Bool.true()
    F = Bool.false()

    p, a, b, c, as_, bs = vars_("p a b c as bs", Bool)
    egraph.register(
        # clauses are assumed in the normal form (or a (or b (or c False)))
        set_(~F).to(T),
        set_(~T).to(F),
        # "Solving" negation equations
        rule(eq(~p).to(T)).then(union(p).with_(F)),
        rule(eq(~p).to(F)).then(union(p).with_(T)),
        # canonicalize associtivity. "append" for clauses terminate with false
        rewrite((a | b) | c).to(a | (b | c)),
        # commutativity
        rewrite(a | (b | c)).to(b | (a | c)),
        # absoprtion
        rewrite(a | (a | b)).to(a | b),
        rewrite(a | (~a | b)).to(T),
        # Simplification
        rewrite(F | a).to(a),
        rewrite(a | F).to(a),
        rewrite(T | a).to(T),
        rewrite(a | T).to(T),
        # unit propagation
        # This is kind of interesting actually.
        # Looks a bit like equation solving
        rule(eq(T).to(p | F)).then(union(p).with_(T)),
        # resolution
        # This counts on commutativity to bubble everything possible up to the front of the clause.
        rule(
            eq(T).to(a | as_),
            eq(T).to(~a | bs),
        ).then(
            set_(as_ | bs).to(T),
        ),
    )

    # Example predicate
    @egraph.function
    def pred(x: i64Like) -> Bool:  # type: ignore[empty-body]
        ...

    p0 = egraph.define("p0", pred(0))
    p1 = egraph.define("p1", pred(1))
    p2 = egraph.define("p2", pred(2))
    egraph.register(
        set_(p1 | (~p2 | F)).to(T),
        set_(p2 | (~p0 | F)).to(T),
        set_(p0 | (~p1 | F)).to(T),
        union(p1).with_(F),
        set_(~p0 | (~p1 | (p2 | F))).to(T),
    )
    egraph.run(10)
    egraph.check(T != F)
    egraph.check(eq(p0).to(F))
    egraph.check(eq(p2).to(F))


def test_push_pop():
    egraph = EGraph()

    @egraph.function(merge=lambda old, new: old.max(new))
    def foo() -> i64:  # type: ignore[empty-body]
        ...

    egraph.register(set_(foo()).to(i64(1)))
    egraph.check(eq(foo()).to(i64(1)))

    with egraph:
        egraph.register(set_(foo()).to(i64(2)))
        egraph.check(eq(foo()).to(i64(2)))

    egraph.check(eq(foo()).to(i64(1)))



import datetime
import json
import pathlib
import subprocess
from typing import Iterable

import pytest
from egg_smol.bindings import *
from egg_smol.commands import parse_and_run


def get_examples_files() -> Iterable[pathlib.Path]:
    """
    Return a list of egg examples files
    """
    metadata_process = subprocess.run(
        ["cargo", "metadata", "--format-version", "1", "-q"],
        capture_output=True,
        check=True,
    )
    metadata = json.loads(metadata_process.stdout)
    (egg_smol_package,) = [
        package for package in metadata["packages"] if package["name"] == "egg-smol"
    ]
    egg_smol_folder = pathlib.Path(egg_smol_package["manifest_path"]).parent
    return (egg_smol_folder / "tests").glob("*.egg")


@pytest.mark.parametrize(
    "example_file", [pytest.param(path, id=path.stem) for path in get_examples_files()]
)
def test_example(example_file: pathlib.Path):
    s = example_file.read_text()
    parse_and_run(s)


class TestEGraph:
    def test_parse_and_run_program(self):
        program = "(check (= (+ 2 2) 4))"
        egraph = EGraph()

        assert egraph.parse_and_run_program(program) == ["Checked."]

    def test_parse_and_run_program_exception(self):
        program = "(check (= 5 4))"
        egraph = EGraph()

        with pytest.raises(
            EggSmolError,
            match='Check failed: Value { tag: "i64", bits: 5 } != Value { tag: "i64", bits: 4 }',
        ):
            egraph.parse_and_run_program(program)

    # These examples are from eqsat-basic
    def test_datatype(self):
        egraph = EGraph()
        egraph.declare_sort("Math")
        egraph.declare_constructor(Variant("Num", ["i64"]), "Math")
        egraph.declare_constructor(Variant("Var", ["String"]), "Math")
        egraph.declare_constructor(Variant("Add", ["Math", "Math"]), "Math")
        egraph.declare_constructor(Variant("Mul", ["Math", "Math"]), "Math")

    def test_define(self):
        egraph = EGraph()
        egraph.declare_sort("Math")
        egraph.declare_constructor(Variant("Num", ["i64"]), "Math")
        egraph.declare_constructor(Variant("Var", ["String"]), "Math")
        egraph.declare_constructor(Variant("Add", ["Math", "Math"]), "Math")
        egraph.declare_constructor(Variant("Mul", ["Math", "Math"]), "Math")

        # (define expr1 (Mul (Num 2) (Add (Var "x") (Num 3))))
        egraph.define(
            "expr1",
            Call(
                "Mul",
                [
                    Call(
                        "Num",
                        [
                            Lit(Int(2)),
                        ],
                    ),
                    Call(
                        "Add",
                        [
                            Call(
                                "Var",
                                [
                                    Lit(String("x")),
                                ],
                            ),
                            Call(
                                "Num",
                                [
                                    Lit(Int(3)),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        )

    def test_rewrite(self):
        egraph = EGraph()
        egraph.declare_sort("Math")
        egraph.declare_constructor(Variant("Add", ["Math", "Math"]), "Math")
        name = egraph.add_rewrite(
            Rewrite(
                Call(
                    "Add",
                    [
                        Var("a"),
                        Var("b"),
                    ],
                ),
                Call(
                    "Add",
                    [
                        Var("b"),
                        Var("a"),
                    ],
                ),
            )
        )
        assert isinstance(name, str)

    def test_run_rules(self):
        egraph = EGraph()
        egraph.declare_sort("Math")
        egraph.declare_constructor(Variant("Add", ["Math", "Math"]), "Math")
        egraph.add_rewrite(
            Rewrite(
                Call(
                    "Add",
                    [
                        Var("a"),
                        Var("b"),
                    ],
                ),
                Call(
                    "Add",
                    [
                        Var("b"),
                        Var("a"),
                    ],
                ),
            )
        )
        start_time = datetime.datetime.now()
        searching, applying, rebuilding = egraph.run_rules(10)
        end_time = datetime.datetime.now()
        assert isinstance(searching, datetime.timedelta)
        assert isinstance(applying, datetime.timedelta)
        assert isinstance(rebuilding, datetime.timedelta)
        total_measured_time = searching + applying + rebuilding
        # Verify  less than the total time (which includes time spent in Python).
        assert total_measured_time < (end_time - start_time)

    def test_check_fact(self):
        egraph = EGraph()
        egraph.declare_sort("Math")
        egraph.declare_constructor(Variant("Num", ["i64"]), "Math")
        egraph.declare_constructor(Variant("Var", ["String"]), "Math")
        egraph.declare_constructor(Variant("Add", ["Math", "Math"]), "Math")
        egraph.declare_constructor(Variant("Mul", ["Math", "Math"]), "Math")

        # (define expr1 (Mul (Num 2) (Add (Var "x") (Num 3))))
        egraph.define(
            "expr1",
            Call(
                "Mul",
                [
                    Call(
                        "Num",
                        [
                            Lit(Int(2)),
                        ],
                    ),
                    Call(
                        "Add",
                        [
                            Call(
                                "Var",
                                [
                                    Lit(String("x")),
                                ],
                            ),
                            Call(
                                "Num",
                                [
                                    Lit(Int(3)),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        )
        # (define expr2 (Add (Num 6) (Mul (Num 2) (Var "x"))))
        egraph.define(
            "expr2",
            Call(
                "Add",
                [
                    Call(
                        "Num",
                        [
                            Lit(Int(6)),
                        ],
                    ),
                    Call(
                        "Mul",
                        [
                            Call(
                                "Num",
                                [
                                    Lit(Int(2)),
                                ],
                            ),
                            Call(
                                "Var",
                                [
                                    Lit(String("x")),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        )
        # (rewrite (Add a b)
        #          (Add b a))
        egraph.add_rewrite(
            Rewrite(
                Call(
                    "Add",
                    [
                        Var("a"),
                        Var("b"),
                    ],
                ),
                Call(
                    "Add",
                    [
                        Var("b"),
                        Var("a"),
                    ],
                ),
            )
        )
        # (rewrite (Mul a (Add b c))
        #          (Add (Mul a b) (Mul a c)))
        egraph.add_rewrite(
            Rewrite(
                Call(
                    "Mul",
                    [
                        Var("a"),
                        Call(
                            "Add",
                            [
                                Var("b"),
                                Var("c"),
                            ],
                        ),
                    ],
                ),
                Call(
                    "Add",
                    [
                        Call(
                            "Mul",
                            [
                                Var("a"),
                                Var("b"),
                            ],
                        ),
                        Call(
                            "Mul",
                            [
                                Var("a"),
                                Var("c"),
                            ],
                        ),
                    ],
                ),
            )
        )

        # (rewrite (Add (Num a) (Num b))
        #          (Num (+ a b)))
        lhs = Call(
            "Add",
            [
                Call(
                    "Num",
                    [
                        Var("a"),
                    ],
                ),
                Call(
                    "Num",
                    [
                        Var("b"),
                    ],
                ),
            ],
        )
        rhs = Call(
            "Num",
            [
                Call(
                    "+",
                    [
                        Var("a"),
                        Var("b"),
                    ],
                )
            ],
        )
        egraph.add_rewrite(Rewrite(lhs, rhs))

        # (rewrite (Mul (Num a) (Num b))
        #          (Num (* a b)))
        lhs = Call(
            "Mul",
            [
                Call(
                    "Num",
                    [
                        Var("a"),
                    ],
                ),
                Call(
                    "Num",
                    [
                        Var("b"),
                    ],
                ),
            ],
        )
        rhs = Call(
            "Num",
            [
                Call(
                    "*",
                    [
                        Var("a"),
                        Var("b"),
                    ],
                )
            ],
        )
        egraph.add_rewrite(Rewrite(lhs, rhs))

        egraph.run_rules(10)
        egraph.check_fact(
            Eq(
                [
                    Var("expr1"),
                    Var("expr2"),
                ]
            )
        )

    def test_extract(self):
        # Example from extraction-cost
        egraph = EGraph()
        egraph.declare_sort("Expr")
        egraph.declare_constructor(Variant("Num", ["i64"], cost=5), "Expr")

        egraph.define("x", Call("Num", [Lit(Int(1))]), cost=10)
        egraph.define("y", Call("Num", [Lit(Int(2))]), cost=1)

        assert egraph.extract_expr(Var("x")) == (6, Call("Num", [Lit(Int(1))]), [])
        assert egraph.extract_expr(Var("y")) == (1, Call("y", []), [])

    def test_extract_string(self):
        egraph = EGraph()
        egraph.define("x", Lit(String("hello")))
        assert egraph.extract_expr(Var("x")) == (0, Lit(String("hello")), [])

    def test_rule(self):
        # Example from fibonacci
        egraph = EGraph()
        egraph.declare_function(FunctionDecl("fib", Schema(["i64"], "i64")))
        egraph.eval_actions(Set("fib", [Lit(Int(0))], Lit(Int(0))))
        egraph.eval_actions(Set("fib", [Lit(Int(1))], Lit(Int(1))))
        egraph.add_rule(
            Rule(
                body=[
                    Eq([Var("f0"), Call("fib", [Var("x")])]),
                    Eq([Var("f1"), Call("fib", [Call("+", [Var("x"), Lit(Int(1))])])]),
                ],
                head=[
                    Set(
                        "fib",
                        [Call("+", [Var("x"), Lit(Int(2))])],
                        Call("+", [Var("f0"), Var("f1")]),
                    ),
                ],
            )
        )
        egraph.run_rules(7)
        egraph.check_fact(Eq([Call("fib", [Lit(Int(7))]), Lit(Int(13))]))

    def test_push_pop(self):
        egraph = EGraph()
        egraph.declare_function(
            FunctionDecl(
                "foo", Schema([], "i64"), merge=Call("max", [Var("old"), Var("new")])
            ),
        )
        egraph.eval_actions(Set("foo", [], Lit(Int(1))))
        egraph.check_fact(Eq([Call("foo", []), Lit(Int(1))]))

        egraph.push()
        egraph.eval_actions(Set("foo", [], Lit(Int(2))))
        egraph.check_fact(Eq([Call("foo", []), Lit(Int(2))]))
        egraph.pop()

        egraph.check_fact(Eq([Call("foo", []), Lit(Int(1))]))

    def test_clear(self):
        egraph = EGraph()
        egraph.define("x", Lit(Int(1)))
        egraph.check_fact(Eq([Call("x", []), Lit(Int(1))]))

        egraph.clear()
        with pytest.raises(EggSmolError):
            egraph.check_fact(Eq([Call("x", []), Lit(Int(1))]))

    def test_clear_rules(self):
        egraph = EGraph()
        egraph.declare_function(FunctionDecl("fib", Schema(["i64"], "i64")))
        egraph.eval_actions(Set("fib", [Lit(Int(0))], Lit(Int(0))))
        egraph.eval_actions(Set("fib", [Lit(Int(1))], Lit(Int(1))))
        egraph.add_rule(
            Rule(
                body=[
                    Eq([Var("f0"), Call("fib", [Var("x")])]),
                    Eq([Var("f1"), Call("fib", [Call("+", [Var("x"), Lit(Int(1))])])]),
                ],
                head=[
                    Set(
                        "fib",
                        [Call("+", [Var("x"), Lit(Int(2))])],
                        Call("+", [Var("f0"), Var("f1")]),
                    ),
                ],
            )
        )
        egraph.clear_rules()
        egraph.run_rules(7)
        with pytest.raises(EggSmolError):
            egraph.check_fact(Eq([Call("fib", [Lit(Int(7))]), Lit(Int(13))]))

    def test_print_size(self):
        egraph = EGraph()
        egraph.declare_function(FunctionDecl("fib", Schema(["i64"], "i64")))
        egraph.eval_actions(Set("fib", [Lit(Int(0))], Lit(Int(0))))
        egraph.eval_actions(Set("fib", [Lit(Int(1))], Lit(Int(1))))
        assert egraph.print_size("fib") == "Function fib has size 2"

    def test_print_function(self):
        egraph = EGraph()
        egraph.declare_function(FunctionDecl("fib", Schema(["i64"], "i64")))
        egraph.eval_actions(Set("fib", [Lit(Int(0))], Lit(Int(0))))
        egraph.eval_actions(Set("fib", [Lit(Int(1))], Lit(Int(1))))
        assert egraph.print_function("fib", 2) == "(fib 0) -> 0\n(fib 1) -> 1\n"

    def test_sort_alias(self):
        # From map example
        egraph = EGraph()
        egraph.declare_sort(
            "MyMap",
            ("Map", [Var("i64"), Var("String")]),
        )
        egraph.define(
            "my_map1",
            Call("insert", [Call("empty", []), Lit(Int(1)), Lit(String("one"))]),
        )
        egraph.define(
            "my_map2",
            Call("insert", [Var("my_map1"), Lit(Int(2)), Lit(String("two"))]),
        )

        egraph.check_fact(
            Eq([Lit(String("one")), Call("get", [Var("my_map1"), Lit(Int(1))])])
        )
        _, expr, _ = egraph.extract_expr(Var("my_map2"))
        assert expr == Call(
            "insert",
            [
                Call("insert", [Call("empty", []), Lit(Int(2)), Lit(String("two"))]),
                Lit(Int(1)),
                Lit(String("one")),
            ],
        )


def test_fib_demand():
    egraph = EGraph()
    # (datatype Expr
    #   (Num i64)
    #   (Add Expr Expr))
    egraph.declare_sort("Expr")
    egraph.declare_constructor(Variant("Num", ["i64"]), "Expr")
    egraph.declare_constructor(Variant("Add", ["Expr", "Expr"]), "Expr")
    # (function Fib (i64) Expr)
    egraph.declare_function(FunctionDecl("Fib", Schema(["i64"], "Expr")))
    # (rewrite (Add (Num a) (Num b)) (Num (+ a b)))
    egraph.add_rewrite(
        Rewrite(
            Call("Add", [Call("Num", [Var("a")]), Call("Num", [Var("b")])]),
            Call("Num", [Call("+", [Var("a"), Var("b")])]),
        )
    )
    # (rule ((= f (Fib x))
    #     (> x 1))
    #     ((set (Fib x) (Add (Fib (- x 1)) (Fib (- x 2))))))
    egraph.add_rule(
        Rule(
            [
                Set(
                    "Fib",
                    [Var("x")],
                    Call(
                        "Add",
                        [
                            Call("Fib", [Call("-", [Var("x"), Lit(Int(1))])]),
                            Call("Fib", [Call("-", [Var("x"), Lit(Int(2))])]),
                        ],
                    ),
                )
            ],
            [
                Eq([Var("f"), Call("Fib", [Var("x")])]),
                Fact(Call(">", [Var("x"), Lit(Int(1))])),
            ],
        )
    )
    # (set (Fib 0) (Num 0))
    egraph.eval_actions(Set("Fib", [Lit(Int(0))], Call("Num", [Lit(Int(0))])))
    # (set (Fib 1) (Num 1))
    egraph.eval_actions(Set("Fib", [Lit(Int(1))], Call("Num", [Lit(Int(1))])))
    # (define f7 (Fib 7))
    egraph.define("f7", Call("Fib", [Lit(Int(7))]))
    # (run 14)
    egraph.run_rules(14)
    # (extract f7)
    egraph.extract_expr(Var("f7"))
    # (check (= f7 (Num 13)))
    egraph.check_fact(Eq([Var("f7"), Call("Num", [Lit(Int(13))])]))


class TestVariant:
    def test_repr(self):
        assert repr(Variant("name", [])) == "Variant('name', [], None)"

    def test_name(self):
        assert Variant("name", []).name == "name"

    def test_types(self):
        assert Variant("name", ["a", "b"]).types == ["a", "b"]

    def test_cost(self):
        assert Variant("name", [], cost=1).cost == 1

    def test_compare(self):
        assert Variant("name", []) == Variant("name", [])
        assert Variant("name", []) != Variant("name", ["a"])
        assert Variant("name", []) != 10  # type: ignore


class TestParse:
    # TODO: Test all examples
    def test_parse_simple(self):
        res = parse(
            """(datatype Math
          (Num i64)
          (Var String)
          (Add Math Math)
          (Mul Math Math))

        ;; expr1 = 2 * (x + 3)
        (define expr1 (Mul (Num 2) (Add (Var "x") (Num 3))))"""
        )

        assert res == [
            Datatype(
                "Math",
                [
                    Variant("Num", ["i64"]),
                    Variant("Var", ["String"]),
                    Variant("Add", ["Math", "Math"]),
                    Variant("Mul", ["Math", "Math"]),
                ],
            ),
            Define(
                "expr1",
                Call(
                    "Mul",
                    [
                        Call("Num", [Lit(Int(2))]),
                        Call(
                            "Add",
                            [
                                Call("Var", [Lit(String('"x"'))]),
                                Call("Num", [Lit(Int(3))]),
                            ],
                        ),
                    ],
                ),
                None,
            ),
        ]



"""
Builtin sorts and function to egg.
"""


from __future__ import annotations

from typing import Generic, TypeVar, Union

from .registry import BUILTINS, BaseExpr, Unit

__all__ = [
    "BUILTINS",
    "i64",
    "i64Like",
    "String",
    "StringLike",
    "Map",
    "Rational",
]


# The types which can be converted into an i64
i64Like = Union[int, "i64"]


@BUILTINS.class_(egg_sort="i64")
class i64(BaseExpr):
    def __init__(self, value: int):
        ...

    @BUILTINS.method(egg_fn="+")
    def __add__(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="-")
    def __sub__(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="*")
    def __mul__(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="/")
    def __truediv__(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="%")
    def __mod__(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="&")
    def __and__(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="|")
    def __or__(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="^")
    def __xor__(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="<<")
    def __lshift__(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn=">>")
    def __rshift__(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="not-64")
    def __invert__(self) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="<")
    def __lt__(self, other: i64Like) -> Unit:  # type: ignore[empty-body,has-type]
        ...

    @BUILTINS.method(egg_fn=">")
    def __gt__(self, other: i64Like) -> Unit:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="min")
    def min(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="max")
    def max(self, other: i64Like) -> i64:  # type: ignore[empty-body]
        ...


@BUILTINS.class_
class String(BaseExpr):
    def __init__(self, value: str):
        ...


StringLike = Union[str, String]

T = TypeVar("T", bound=BaseExpr)
V = TypeVar("V", bound=BaseExpr)


@BUILTINS.class_(egg_sort="Map")
class Map(BaseExpr, Generic[T, V]):
    @BUILTINS.method(egg_fn="empty")
    @classmethod
    def empty(cls) -> Map[T, V]:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="insert")
    def insert(self, key: T, value: V) -> Map[T, V]:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="get")
    def __getitem__(self, key: T) -> V:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="not-contains")
    def not_contains(self, key: T) -> Unit:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="contains")
    def contains(self, key: T) -> Unit:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="set-union")
    def __or__(self, __t: Map[T, V]) -> Map[T, V]:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="set-diff")
    def __sub__(self, __t: Map[T, V]) -> Map[T, V]:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="set-intersect")
    def __and__(self, __t: Map[T, V]) -> Map[T, V]:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="map-remove")
    def map_remove(self, key: T) -> Map[T, V]:  # type: ignore[empty-body]
        ...


@BUILTINS.class_(egg_sort="Rational")
class Rational(BaseExpr):
    @BUILTINS.method(egg_fn="rational")
    def __init__(self, num: i64Like, den: i64Like):
        ...

    @BUILTINS.method(egg_fn="+")
    def __add__(self, other: Rational) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="-")
    def __sub__(self, other: Rational) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="*")
    def __mul__(self, other: Rational) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="/")
    def __truediv__(self, other: Rational) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="min")
    def min(self, other: Rational) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="max")
    def max(self, other: Rational) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="neg")
    def __neg__(self) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="abs")
    def __abs__(self) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="floor")
    def floor(self) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="ceil")
    def ceil(self) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="round")
    def round(self) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="pow")
    def __pow__(self, other: Rational) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="log")
    def log(self) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="sqrt")
    def sqrt(self) -> Rational:  # type: ignore[empty-body]
        ...

    @BUILTINS.method(egg_fn="cbrt")
    def cbrt(self) -> Rational:  # type: ignore[empty-body]
        ...



from datetime import timedelta

from typing_extensions import final

def parse(input: str) -> list[_Command]: ...
@final
class EGraph:
    def parse_and_run_program(self, input: str) -> list[str]: ...
    def declare_constructor(self, variant: Variant, sort: str) -> None: ...
    def declare_sort(
        self, name: str, presort_and_args: tuple[str, list[_Expr]] | None = None
    ) -> None: ...
    def declare_function(self, decl: FunctionDecl) -> None: ...
    def define(self, name: str, expr: _Expr, cost: int | None = None) -> None: ...
    def add_rewrite(self, rewrite: Rewrite) -> str: ...
    def run_rules(self, limit: int) -> tuple[timedelta, timedelta, timedelta]: ...
    def check_fact(self, fact: _Fact) -> None: ...
    def extract_expr(
        self, expr: _Expr, variants: int = 0
    ) -> tuple[int, _Expr, list[_Expr]]: ...
    def add_rule(self, rule: Rule) -> str: ...
    def eval_actions(self, *actions: _Action) -> None: ...
    def push(self) -> None: ...
    def pop(self) -> None: ...
    def clear(self) -> None: ...
    def clear_rules(self) -> None: ...
    def print_size(self, name: str) -> str: ...
    def print_function(self, name: str, n: int) -> str: ...

@final
class EggSmolError(Exception):
    context: str

@final
class Int:
    def __init__(self, value: int) -> None: ...
    value: int

@final
class String:
    def __init__(self, value: str) -> None: ...
    value: str

@final
class Unit:
    def __init__(self) -> None: ...

_Literal = Int | String | Unit

@final
class Lit:
    def __init__(self, value: _Literal) -> None: ...
    value: _Literal

@final
class Var:
    def __init__(self, name: str) -> None: ...
    name: str

@final
class Call:
    def __init__(self, name: str, args: list[_Expr]) -> None: ...
    name: str
    args: list[_Expr]

# This must be private becuase it is not actually exposed by the runtime library.
_Expr = Lit | Var | Call

@final
class Eq:
    def __init__(self, exprs: list[_Expr]) -> None: ...
    exprs: list[_Expr]

@final
class Fact:
    def __init__(self, expr: _Expr) -> None: ...
    expr: _Expr

_Fact = Fact | Eq

@final
class Let:
    def __init__(self, lhs: str, rhs: _Expr) -> None: ...
    lhs: str
    rhs: _Expr

@final
class Set:
    def __init__(self, lhs: str, args: list[_Expr], rhs: _Expr) -> None: ...
    lhs: str
    args: list[_Expr]
    rhs: _Expr

@final
class Delete:
    sym: str
    args: list[_Expr]
    def __init__(self, sym: str, args: list[_Expr]) -> None: ...

@final
class Union:
    def __init__(self, lhs: _Expr, rhs: _Expr) -> None: ...
    lhs: _Expr
    rhs: _Expr

@final
class Panic:
    def __init__(self, msg: str) -> None: ...
    msg: str

@final
class Expr_:
    def __init__(self, expr: _Expr) -> None: ...
    expr: _Expr

_Action = Let | Set | Delete | Union | Panic | Expr_

@final
class FunctionDecl:
    name: str
    schema: Schema
    default: _Expr | None
    merge: _Expr | None
    cost: int | None
    def __init__(
        self,
        name: str,
        schema: Schema,
        default: _Expr | None = None,
        merge: _Expr | None = None,
        cost: int | None = None,
    ) -> None: ...

@final
class Variant:
    def __init__(
        self, name: str, types: list[str], cost: int | None = None
    ) -> None: ...
    name: str
    types: list[str]
    cost: int | None

@final
class Schema:
    input: list[str]
    output: str
    def __init__(self, input: list[str], output: str) -> None: ...

@final
class Rule:
    head: list[_Action]
    body: list[_Fact]
    def __init__(self, head: list[_Action], body: list[_Fact]) -> None: ...

@final
class Rewrite:
    lhs: _Expr
    rhs: _Expr
    conditions: list[_Fact]

    def __init__(
        self, lhs: _Expr, rhs: _Expr, conditions: list[_Fact] = []
    ) -> None: ...

@final
class Datatype:
    name: str
    variants: list[Variant]
    def __init__(self, name: str, variants: list[Variant]) -> None: ...

@final
class Sort:
    name: str
    presort: str
    args: list[_Expr]
    def __init__(self, name: str, presort: str, args: list[_Expr]) -> None: ...

@final
class Function:
    decl: FunctionDecl
    def __init__(self, decl: FunctionDecl) -> None: ...

@final
class Define:
    name: str
    expr: _Expr
    cost: int | None
    def __init__(self, name: str, expr: _Expr, cost: int | None = None) -> None: ...

@final
class RuleCommand:
    rule: Rule
    def __init__(self, rule: Rule) -> None: ...

@final
class RewriteCommand:
    rewrite: Rewrite
    def __init__(self, rewrite: Rewrite) -> None: ...

@final
class ActionCommand:
    action: _Action
    def __init__(self, action: _Action) -> None: ...

@final
class Run:
    length: int
    def __init__(self, length: int) -> None: ...

@final
class Extract:
    variants: int
    expr: _Expr
    def __init__(self, variants: int, expr: _Expr) -> None: ...

@final
class Check:
    fact: _Fact
    def __init__(self, fact: _Fact) -> None: ...

@final
class ClearRules:
    def __init__(self) -> None: ...

@final
class Clear:
    def __init__(self) -> None: ...

@final
class Print:
    name: str
    length: int
    def __init__(self, name: str, length: int) -> None: ...

@final
class PrintSize:
    name: str
    def __init__(self, name: str) -> None: ...

@final
class Input:
    name: str
    file: str
    def __init__(self, name: str, file: str) -> None: ...

@final
class Query:
    facts: list[_Fact]
    def __init__(self, facts: list[_Fact]) -> None: ...

@final
class Push:
    length: int
    def __init__(self, length: int) -> None: ...

@final
class Pop:
    length: int
    def __init__(self, length: int) -> None: ...

_Command = (
    Datatype
    | Sort
    | Function
    | Define
    | RuleCommand
    | RewriteCommand
    | ActionCommand
    | Run
    | Extract
    | Check
    | ClearRules
    | Clear
    | Print
    | PrintSize
    | Input
    | Query
    | Push
    | Pop
)



"""
This module holds a number of types which are only used at runtime to emulate Python objects.

Users will not import anything from this module, and statically they won't know these are the types they are using.

But at runtime they will be exposed.

Note that all their internal fields are prefixed with __egg_ to avoid name collisions with user code.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Collection, Iterable, Optional, Union

import black
from typing_extensions import assert_never

from . import config
from .declarations import *
from .declarations import BINARY_METHODS, UNARY_METHODS
from .type_constraint_solver import *

__all__ = [
    "LIT_CLASS_NAMES",
    "RuntimeClass",
    "RuntimeParamaterizedClass",
    "RuntimeClassMethod",
    "RuntimeExpr",
    "RuntimeFunction",
    "ArgType",
]


BLACK_MODE = black.Mode(line_length=120)  # type: ignore

UNIT_CLASS_NAME = "unit"
UNARY_LIT_CLASS_NAMES = {"i64", "String"}
LIT_CLASS_NAMES = UNARY_LIT_CLASS_NAMES | {UNIT_CLASS_NAME}


@dataclass
class RuntimeClass:
    __egg_decls__: Declarations
    __egg_name__: str

    def __call__(self, *args: ArgType) -> RuntimeExpr:
        """
        Create an instance of this kind by calling the __init__ classmethod
        """
        # If this is a literal type, initializing it with a literal should return a literal
        if self.__egg_name__ in UNARY_LIT_CLASS_NAMES:
            assert len(args) == 1
            assert isinstance(args[0], (int, str))
            return RuntimeExpr(
                self.__egg_decls__, JustTypeRef(self.__egg_name__), LitDecl(args[0])
            )
        if self.__egg_name__ == UNIT_CLASS_NAME:
            assert len(args) == 0
            return RuntimeExpr(
                self.__egg_decls__, JustTypeRef(self.__egg_name__), LitDecl(None)
            )

        return RuntimeClassMethod(self.__egg_decls__, self.__egg_name__, "__init__")(
            *args
        )

    def __dir__(self) -> list[str]:
        possible_methods = list(
            self.__egg_decls__.classes[self.__egg_name__].class_methods
        )
        if "__init__" in possible_methods:
            possible_methods.remove("__init__")
            possible_methods.append("__call__")
        return possible_methods

    def __getitem__(
        self, args: tuple[RuntimeTypeArgType, ...]
    ) -> RuntimeParamaterizedClass:
        tp = JustTypeRef(self.__egg_name__, tuple(class_to_ref(arg) for arg in args))
        return RuntimeParamaterizedClass(self.__egg_decls__, tp)

    def __getattr__(self, name: str) -> RuntimeClassMethod:
        return RuntimeClassMethod(self.__egg_decls__, self.__egg_name__, name)

    def __str__(self) -> str:
        return self.__egg_name__

    # Make hashable so can go in Union
    def __hash__(self) -> int:
        return hash((id(self.__egg_decls__), self.__egg_name__))


@dataclass
class RuntimeParamaterizedClass:
    __egg_decls__: Declarations
    # Note that this will never be a typevar because we don't use RuntimeParamaterizedClass for maps on their own methods
    # which is the only time we define function which take typevars
    __egg_tp__: JustTypeRef

    def __post_init__(self):
        desired_args = self.__egg_decls__.classes[self.__egg_tp__.name].n_type_vars
        if len(self.__egg_tp__.args) != desired_args:
            raise ValueError(
                f"Expected {desired_args} type args, got {len(self.__egg_tp__.args)}"
            )

    def __call__(self, *args: ArgType) -> RuntimeExpr:
        return RuntimeClassMethod(self.__egg_decls__, class_to_ref(self), "__init__")(
            *args
        )

    def __getattr__(self, name: str) -> RuntimeClassMethod:
        return RuntimeClassMethod(self.__egg_decls__, class_to_ref(self), name)

    def __str__(self) -> str:
        return self.__egg_tp__.pretty()


# Type args can either be typevars or classes
RuntimeTypeArgType = Union[RuntimeClass, RuntimeParamaterizedClass]


def class_to_ref(cls: RuntimeTypeArgType) -> JustTypeRef:
    if isinstance(cls, RuntimeClass):
        return JustTypeRef(cls.__egg_name__)
    if isinstance(cls, RuntimeParamaterizedClass):
        return cls.__egg_tp__
    assert_never(cls)


@dataclass
class RuntimeFunction:
    __egg_decls__: Declarations
    __egg_name__: str

    def __post_init__(self):
        if self.__egg_name__ not in self.__egg_decls__.functions:
            raise ValueError(f"Function {self.__egg_name__} does not exist")

    def __call__(self, *args: ArgType) -> RuntimeExpr:
        return _call(
            self.__egg_decls__,
            FunctionRef(self.__egg_name__),
            self.__egg_decls__.functions[self.__egg_name__],
            args,
        )

    def __str__(self) -> str:
        return self.__egg_name__


def _call(
    decls: Declarations,
    callable_ref: CallableRef,
    # Not included if this is the != method
    fn_decl: Optional[FunctionDecl],
    args: Collection[ArgType],
    bound_params: Optional[tuple[JustTypeRef, ...]] = None,
) -> RuntimeExpr:
    upcasted_args = [_resolve_literal(decls, arg) for arg in args]

    arg_types = [arg.__egg_tp__ for arg in upcasted_args]

    if bound_params is not None:
        tcs = TypeConstraintSolver.from_type_parameters(bound_params)
    else:
        tcs = TypeConstraintSolver()

    if fn_decl is not None:
        return_tp = tcs.infer_return_type(
            fn_decl.arg_types, fn_decl.return_type, arg_types
        )
    else:
        return_tp = JustTypeRef("unit")

    arg_decls = tuple(arg.__egg_expr__ for arg in upcasted_args)
    expr_decl = CallDecl(callable_ref, arg_decls, bound_params)
    return RuntimeExpr(decls, return_tp, expr_decl)


@dataclass
class RuntimeClassMethod:
    __egg_decls__: Declarations
    # Either a string if it isn't bound or a tp if it s
    __egg_tp__: JustTypeRef | str
    __egg_method_name__: str

    def __post_init__(self):
        if (
            self.__egg_method_name__
            not in self.__egg_decls__.classes[self.class_name].class_methods
        ):
            raise AttributeError(
                f"Class {self.class_name} does not have method {self.__egg_method_name__}"
            )

    def __call__(self, *args: ArgType) -> RuntimeExpr:
        fn_decl = self.__egg_decls__.classes[self.class_name].class_methods[
            self.__egg_method_name__
        ]
        bound_params = (
            self.__egg_tp__.args if isinstance(self.__egg_tp__, JustTypeRef) else None
        )
        return _call(
            self.__egg_decls__,
            ClassMethodRef(self.class_name, self.__egg_method_name__),
            fn_decl,
            args,
            bound_params,
        )

    def __str__(self) -> str:
        return f"{self.class_name}.{self.__egg_method_name__}"

    @property
    def class_name(self) -> str:
        if isinstance(self.__egg_tp__, str):
            return self.__egg_tp__
        return self.__egg_tp__.name


@dataclass
class RuntimeMethod:
    __egg_decls__: Declarations
    __egg_tp__: JustTypeRef
    __egg_method_name__: str
    __egg_slf_arg__: ExprDecl

    def __post_init__(self):
        if (
            self.__egg_method_name__
            not in self.__egg_decls__.classes[self.class_name].methods
            # Special case for __ne__ which does not have a normal function defintion since
            # it relies of type parameters
            and self.__egg_method_name__ != "__ne__"
        ):
            raise AttributeError(
                f"Class {self.class_name} does not have method {self.__egg_method_name__}"
            )

    def __call__(self, *args: ArgType) -> RuntimeExpr:
        fn_decl = (
            self.__egg_decls__.classes[self.class_name].methods[
                self.__egg_method_name__
            ]
            if self.__egg_method_name__ != "__ne__"
            else None
        )

        first_arg = RuntimeExpr(
            self.__egg_decls__, JustTypeRef(self.class_name), self.__egg_slf_arg__
        )
        args = (first_arg, *args)

        return _call(
            self.__egg_decls__,
            MethodRef(self.class_name, self.__egg_method_name__),
            fn_decl,
            args,
        )

    @property
    def class_name(self) -> str:
        return self.__egg_tp__.name


@dataclass
class RuntimeExpr:
    __egg_decls__: Declarations
    __egg_tp__: JustTypeRef
    __egg_expr__: ExprDecl

    @property
    def __egg_parts__(self) -> tuple[JustTypeRef, ExprDecl]:
        return self.__egg_tp__, self.__egg_expr__

    def __getattr__(self, name: str) -> RuntimeMethod:
        return RuntimeMethod(
            self.__egg_decls__, self.__egg_tp__, name, self.__egg_expr__
        )

    def __repr__(self) -> str:
        """
        The repr of the expr is the pretty printed version of the expr.
        """
        return str(self)

    def __str__(self) -> str:
        pretty_expr = self.__egg_expr__.pretty(parens=False)
        if config.SHOW_TYPES:
            s = f"_: {self.__egg_tp__.pretty()} = {pretty_expr}"
            return black.format_str(s, mode=black.FileMode()).strip()
        else:
            return black.format_str(
                pretty_expr, mode=black.FileMode(line_length=180)
            ).strip()

    def __dir__(self) -> Iterable[str]:
        return list(self.__egg_decls__.classes[self.__egg_tp__.name].methods)

    # Have __eq__ take no NoReturn (aka Never https://docs.python.org/3/library/typing.html#typing.Never) because
    # we don't wany any type that MyPy thinks is an expr to be used with __eq__.
    # That's because we want to reserve __eq__ for domain specific equality checks, overloading this method.
    # To check if two exprs are equal, use the expr_eq method.
    def __eq__(self, other: NoReturn) -> Expr:  # type: ignore
        raise NotImplementedError("Compare the __parts__ attribute instead")


# Define each of the special methods, since we have already declared them for pretty printing
for name in list(BINARY_METHODS) + list(UNARY_METHODS) + ["__getitem__", "__call__"]:

    def _special_method(
        self: RuntimeExpr, *args: ArgType, __name: str = name
    ) -> RuntimeExpr:
        return RuntimeMethod(
            self.__egg_decls__, self.__egg_tp__, __name, self.__egg_expr__
        )(*args)

    setattr(RuntimeExpr, name, _special_method)


# Args can either be expressions or literals which are automatically promoted
ArgType = Union[RuntimeExpr, int, str]

tp_to_lit: dict[type, str] = {
    int: "i64",
    str: "String",
}


def _resolve_literal(decls: Declarations, arg: ArgType) -> RuntimeExpr:
    if isinstance(arg, int):
        return RuntimeExpr(decls, JustTypeRef("i64"), LitDecl(arg))
    elif isinstance(arg, str):
        return RuntimeExpr(decls, JustTypeRef("String"), LitDecl(arg))
    return arg


def test_type_str():
    decls = Declarations(
        classes={
            "i64": ClassDecl(),
            "Map": ClassDecl(n_type_vars=2),
        }
    )
    i64 = RuntimeClass(decls, "i64")
    Map = RuntimeClass(decls, "Map")
    assert str(i64) == "i64"
    assert str(Map[i64, i64]) == "Map[i64, i64]"


def test_function_call():
    decls = Declarations(
        classes={
            "i64": ClassDecl(),
        },
        functions={
            "one": FunctionDecl(
                (),
                TypeRefWithVars("i64"),
            ),
        },
    )
    one = RuntimeFunction(decls, "one")
    assert (
        one().__egg_parts__
        == RuntimeExpr(
            decls, JustTypeRef("i64"), CallDecl(FunctionRef("one"))
        ).__egg_parts__
    )


def test_classmethod_call():
    from pytest import raises

    K, V = ClassTypeVarRef(0), ClassTypeVarRef(1)
    decls = Declarations(
        classes={
            "i64": ClassDecl(),
            "unit": ClassDecl(),
            "Map": ClassDecl(n_type_vars=2),
        },
        functions={
            "create": FunctionDecl(
                (),
                TypeRefWithVars("Map", (K, V)),
            )
        },
    )
    Map = RuntimeClass(decls, "Map")
    with raises(TypeConstraintError):
        Map.create()
    i64 = RuntimeClass(decls, "i64")
    unit = RuntimeClass(decls, "unit")
    assert (
        Map[i64, unit].create().__egg_parts__
        == RuntimeExpr(
            decls,
            JustTypeRef("Map", (JustTypeRef("i64"), JustTypeRef("unit"))),
            CallDecl(
                ClassMethodRef("Map", "create"),
                (),
                (JustTypeRef("i64"), JustTypeRef("unit")),
            ),
        ).__egg_parts__
    )


def test_expr_special():
    decls = Declarations(
        classes={
            "i64": ClassDecl(
                methods={
                    "__add__": FunctionDecl(
                        (TypeRefWithVars("i64"), TypeRefWithVars("i64")),
                        TypeRefWithVars("i64"),
                    )
                },
                class_methods={
                    "__init__": FunctionDecl(
                        (TypeRefWithVars("i64"),),
                        TypeRefWithVars("i64"),
                    )
                },
            ),
        },
    )
    i64 = RuntimeClass(decls, "i64")
    one = i64(1)  # type: ignore
    res = one + one  # type: ignore
    expected_res = RuntimeExpr(
        decls,
        JustTypeRef("i64"),
        CallDecl(MethodRef("i64", "__add__"), (LitDecl(1), LitDecl(1))),
    )
    assert res.parts == expected_res.__egg_parts__


# def blacken_python_expression(expr: str) -> str:
#     """
#     Runs black on a Python expression, to remove excess paranthesis and wrap it.
#     """
#     return black.format_str("x = " + expr, mode=BLACK_MODE)[4:-1]
